import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.fci import direct_spin1
from pyscf.mcscf import newton_casscf
from pyscf.grad import casscf as casscf_grad
from pyscf.grad import sacasscf as sacasscf_grad
from functools import reduce

# The extension from gradients -> NACs has three basic steps:
# 0. ("state" index integer -> tuple)
# 1. fcisolver.make_rdm12 -> fcisolver.trans_rdm12
# 2. remove core-orbital and nuclear contributions to everything
# 3. option to include the "csf contribution"
# Additional good ideas:
# a. Option to multiply NACs by the energy difference to control
#    singularities

def _unpack_state(state):
    assert len(state) == 2, "derivative couplings are defined between 2 states"
    return state[0], state[1]


def grad_elec_core(mc_grad, mo_coeff=None, atmlst=None, eris=None, mf_grad=None):
    """Compute the core-electron part of the CASSCF (Hellmann-Feynman)
    gradient using a modified RHF grad_elec call."""
    mc = mc_grad.base
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if eris is None: eris = mc.ao2mo (mo_coeff)
    if mf_grad is None: mf_grad = mc._scf.nuc_grad_method ()
    ncore = mc.ncore
    moH = mo_coeff.conj ().T
    f0 = (moH @ mc.get_hcore () @ mo_coeff) + eris.vhf_c
    mo_energy = f0.diagonal ().copy ()
    mo_occ = np.zeros_like (mo_energy)
    mo_occ[:ncore] = 2.0
    f0 *= mo_occ[None,:]
    dme0 = lambda * args: mo_coeff @ ((f0+f0.T)*.5) @ moH
    with lib.temporary_env (mf_grad, make_rdm1e=dme0, verbose=0):
        with lib.temporary_env (mf_grad.base, mo_coeff=mo_coeff, mo_occ=mo_occ):
            # Second level there should become unnecessary in future, if anyone
            # ever gets around to cleaning up pyscf.df.grad.rhf & pyscf.grad.rhf
            de = mf_grad.grad_elec (mo_coeff=mo_coeff, mo_energy=mo_energy,
                                    mo_occ=mo_occ, atmlst=atmlst)
    return de

def grad_elec_active (mc_grad, mo_coeff=None, ci=None, atmlst=None,
                      eris=None, mf_grad=None, verbose=None):
    '''Compute the active-electron part of the CASSCF (Hellmann-Feynman)
    gradient by subtracting the core-electron part.'''
    t0 = (logger.process_clock (), logger.perf_counter ())
    mc = mc_grad.base
    log = logger.new_logger (mc_grad, verbose)
    if mf_grad is None: mf_grad=mc._scf.nuc_grad_method ()
    de = mc_grad.grad_elec (mo_coeff=mo_coeff, ci=ci, atmlst=atmlst,
                            verbose=0)
    de -= grad_elec_core (mc_grad, mo_coeff=mo_coeff, atmlst=atmlst,
                          eris=eris, mf_grad=mf_grad)
    log.debug ('CASSCF active-orbital gradient:\n{}'.format (de))
    log.timer ('CASSCF active-orbital gradient', *t0)
    return de

def gen_g_hop_active (mc, mo, ci0, eris, verbose=None):
    '''Compute the active-electron part of the orbital rotation gradient
    by patching out the appropriate block of eris.vhf_c'''
    moH = mo.conj ().T
    ncore = mc.ncore
    vnocore = eris.vhf_c.copy ()
    vnocore[:,:ncore] = -moH @ mc.get_hcore () @ mo[:,:ncore]
    with lib.temporary_env (eris, vhf_c=vnocore):
        return newton_casscf.gen_g_hop (mc, mo, ci0, eris, verbose=verbose)

def _nac_csf (mol, mf_grad, tm1, atmlst):
    if atmlst is None: atmlst = list (range (mol.natm))
    aoslices = mol.aoslice_by_atom ()
    s1 = mf_grad.get_ovlp (mol)
    # if libcint documentation is to be trusted, mf_grad.get_ovlp
    # corresponds to differentiating on the SECOND index: <p|dq/dR>
    nac = np.zeros ((len(atmlst), 3))
    for k, ia in enumerate (atmlst):
        shl0, shl1, p0, p1 = aoslices[ia]
        nac[k] += 0.5*np.einsum ('xij,ij->x', s1[:,p0:p1], tm1[p0:p1])
    return nac

def nac_csf (mc_grad, mo_coeff=None, ci=None, state=None, mf_grad=None,
             atmlst=None):
    '''Compute the "CSF contribution" to the SA-CASSCF NAC'''
    mc = mc_grad.base
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if state is None: state = mc_grad.state
    if mf_grad is None: mf_grad = mc._scf.nuc_grad_method ()
    if atmlst is None: atmlst = mc_grad.atmlst
    mol = mc.mol
    ket, bra = _unpack_state (state)
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    castm1 = direct_spin1.trans_rdm1 (ci[bra], ci[ket], ncas, nelecas)
    # if PySCF commentary is to be trusted, trans_rdm1[p,q] is
    # <bra|q'p|ket>. I want <bra|p'q - q'p|ket>.
    castm1 = castm1.conj ().T - castm1
    mo_cas = mo_coeff[:,ncore:][:,:ncas]
    tm1 = reduce (np.dot, (mo_cas, castm1, mo_cas.conj ().T))
    return _nac_csf (mol, mf_grad, tm1, atmlst)

class NonAdiabaticCouplings (sacasscf_grad.Gradients):
    '''SA-CASSCF non-adiabatic couplings (NACs) between states

    kwargs/attributes:

    state : tuple of length 2
        The NACs returned are <state[1]|d(state[0])/dR>.
        In other words, state = (ket, bra).
    mult_ediff : logical
        If True, returns NACs multiplied by the energy difference.
        Useful near conical intersections to avoid numerical problems.
    use_etfs : logical
        If True, use the ``electron translation factors'' of Fatehi and
        Subotnik [JPCL 3, 2039 (2012)], which guarantee conservation of
        total electron + nuclear momentum when the nuclei are moving
        (i.e., in non-adiabatic molecular dynamics). This corresponds
        to omitting the so-called ``CSF contribution'' [cf. JCTC 12,
        3636 (2016)].
    '''

    def __init__(self, mc, state=None, mult_ediff=False, use_etfs=False):
        self.mult_ediff = mult_ediff
        self.use_etfs = use_etfs
        if state is not None:
            assert len(state) == 2, "derivative couplings are defined between 2 states"
        sacasscf_grad.Gradients.__init__(self, mc, state=state)

    def make_fcasscf_nacs (self, state=None, casscf_attr=None,
                           fcisolver_attr=None):
        if state is None: state = self.state
        if casscf_attr is None: casscf_attr = {}
        if fcisolver_attr is None: fcisolver_attr = {}
        ket, bra = _unpack_state (state)
        ci, ncas, nelecas = self.base.ci, self.base.ncas, self.base.nelecas
        # TODO: use fcisolver.fcisolvers in state-average mix case for this
        castm1, castm2 = direct_spin1.trans_rdm12 (ci[bra], ci[ket], ncas,
                                                   nelecas)
        castm1 = 0.5 * (castm1 + castm1.T)
        castm2 = 0.5 * (castm2 + castm2.transpose (1,0,3,2))
        fcisolver_attr['make_rdm12'] = lambda *args, **kwargs : (castm1, castm2)
        fcisolver_attr['make_rdm1'] = lambda *args, **kwargs : castm1
        fcisolver_attr['make_rdm2'] = lambda *args, **kwargs : castm2
        return sacasscf_grad.Gradients.make_fcasscf (self,
            state=ket, casscf_attr=casscf_attr, fcisolver_attr=fcisolver_attr)


    def get_wfn_response (self, atmlst=None, state=None, verbose=None, mo=None, ci=None, **kwargs):
        if state is None: state = self.state
        if atmlst is None: atmlst = self.atmlst
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        log = logger.new_logger (self, verbose)
        ket, bra = _unpack_state (state)
        fcasscf = self.make_fcasscf_nacs (state)
        fcasscf.mo_coeff = mo
        fcasscf.ci = ci[ket]
        eris = fcasscf.ao2mo (mo)
        g_all_ket = gen_g_hop_active (fcasscf, mo, ci[ket], eris, verbose)[0]
        g_all = np.zeros (self.nlag)
        g_all[:self.ngorb] = g_all_ket[:self.ngorb]
        # The fun thing about the ci sector is that you swap them (&/2):
        # <I|[H,|A><I|-|I><A|]|J> = <A|H|J> = <J|[H,|A><J|-|J><A|]|J>/2
        # (It should be zero for converged SA-CASSCF anyway, though)
        g_ci_bra = 0.5 * g_all_ket[self.ngorb:]
        g_all_bra = gen_g_hop_active (fcasscf, mo, ci[bra], eris, verbose)[0]
        g_ci_ket = 0.5 * g_all_bra[self.ngorb:]
        # I have to make sure they don't talk to each other because the
        # preconditioner doesn't explore that space at all. Should I
        # instead solve at the init_guess step, like in MC-PDFT?
        # In practice it should all be zeros but how tightly does
        # everything have to be converged?
        ndet_ket = (self.na_states[ket], self.nb_states[ket])
        ndet_bra = (self.na_states[bra], self.nb_states[bra])
        if ndet_ket==ndet_bra:
            ket2bra = np.dot (ci[bra].conj ().ravel (), g_ci_ket)
            bra2ket = np.dot (ci[ket].conj ().ravel (), g_ci_bra)
            log.debug ('SA-CASSCF <bra|H|ket>,<ket|H|bra> check: %5.3g , %5.3g',
                       ket2bra, bra2ket)
            g_ci_ket -= ket2bra * ci[bra].ravel ()
            g_ci_bra -= bra2ket * ci[ket].ravel ()
        ndet_ket = ndet_ket[0]*ndet_ket[1]
        ndet_bra = ndet_bra[0]*ndet_bra[1]
        # No need to reshape or anything, just use the magic of repeated slicing
        offs_ket = (sum ([na * nb for na, nb in zip(
                         self.na_states[:ket], self.nb_states[:ket])])
                    if ket > 0 else 0)
        offs_bra = (sum ([na * nb for na, nb in zip(
                         self.na_states[:bra], self.nb_states[:bra])])
                    if ket > 0 else 0)
        g_all[self.ngorb:][offs_ket:][:ndet_ket] = g_ci_ket
        g_all[self.ngorb:][offs_bra:][:ndet_bra] = g_ci_bra
        return g_all


    def get_ham_response (self, state=None, atmlst=None, verbose=None, mo=None,
                          ci=None, eris=None, mf_grad=None, **kwargs):
        if state is None: state = self.state
        if atmlst is None: atmlst = self.atmlst
        if verbose is None: verbose = self.verbose
        if mo is None: mo = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if mf_grad is None: mf_grad = self.base._scf.nuc_grad_method ()
        if eris is None and self.eris is None:
            eris = self.eris = self.base.ao2mo (mo)
        elif eris is None:
            eris = self.eris
        use_etfs = kwargs.get ('use_etfs', self.use_etfs)
        ket, bra = _unpack_state (state)
        fcasscf_grad = casscf_grad.Gradients (self.make_fcasscf_nacs (state))
        nac = grad_elec_active (fcasscf_grad, mo_coeff=mo, ci=ci[ket],
                                eris=eris, atmlst=atmlst, verbose=verbose)
        if not use_etfs: nac += self.nac_csf (
            mo_coeff=mo, ci=ci, state=state, mf_grad=mf_grad, atmlst=atmlst)
        return nac

    def nac_csf (self, mo_coeff=None, ci=None, state=None, mf_grad=None, atmlst=None):
        if state is None: state = self.state
        if atmlst is None: atmlst = self.atmlst
        if mo_coeff is None: mo_coeff = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if mf_grad is None: mf_grad = self.base._scf.nuc_grad_method ()
        nac = nac_csf (self, mo_coeff=mo_coeff, ci=ci, state=state,
                       mf_grad=mf_grad, atmlst=atmlst)
        ket, bra = _unpack_state (state)
        e_bra = self.base.e_states[bra]
        e_ket = self.base.e_states[ket]
        nac *= e_bra - e_ket
        return nac

    def kernel (self, *args, **kwargs):
        mult_ediff = kwargs.get ('mult_ediff', self.mult_ediff)
        state = kwargs.get ('state', self.state)
        assert len(state) == 2, "derivative couplings are defined between 2 states"
        if state[0] == state[1]:
            mol = kwargs.get('mol', self.mol)
            atmlst = kwargs.get('atmlst', range(mol.natm))
            return np.zeros((len(atmlst), 3))

        nac = sacasscf_grad.Gradients.kernel (self, *args, **kwargs)
        if not mult_ediff:
            ket, bra = _unpack_state (state)
            e_bra = self.base.e_states[bra]
            e_ket = self.base.e_states[ket]
            nac /= e_bra - e_ket
        return nac

if __name__=='__main__':
    from pyscf import gto, scf, mcscf
    from scipy import linalg
    mol = gto.M (atom = 'Li 0 0 0; H 0 0 1.5', basis='sto-3g',
                 output='sacasscf_nacs.log', verbose=lib.logger.INFO)
    mf = scf.RHF (mol).run ()
    mc = mcscf.CASSCF (mf, 2, 2).fix_spin_(ss=0).state_average ([0.5,0.5]).run (conv_tol=1e-10)
    openmolcas_energies = np.array ([-7.85629118, -7.72175252])
    print ("energies:",mc.e_states)
    print ("disagreement w openmolcas:", np.around (mc.e_states-openmolcas_energies, 8))
    mc_nacs = NonAdiabaticCouplings (mc)
    print ("no csf contr")
    nac_01 = mc_nacs.kernel (state=(0,1), use_etfs=True)
    nac_10 = mc_nacs.kernel (state=(1,0), use_etfs=True)
    nac_01_mult = mc_nacs.kernel (state=(0,1), use_etfs=True, mult_ediff=True)
    nac_10_mult = mc_nacs.kernel (state=(1,0), use_etfs=True, mult_ediff=True)
    print ("antisym")
    print (nac_01)
    print ("checking antisym:",linalg.norm(nac_01+nac_10))
    print ("sym")
    print (nac_01_mult)
    print ("checking sym:",linalg.norm(nac_01_mult-nac_10_mult))


    print ("incl csf contr")
    nac_01 = mc_nacs.kernel (state=(0,1), use_etfs=False)
    nac_10 = mc_nacs.kernel (state=(1,0), use_etfs=False)
    nac_01_mult = mc_nacs.kernel (state=(0,1), use_etfs=False, mult_ediff=True)
    nac_10_mult = mc_nacs.kernel (state=(1,0), use_etfs=False, mult_ediff=True)
    print ("antisym")
    print (nac_01)
    print ("checking antisym:",linalg.norm(nac_01+nac_10))
    print ("sym")
    print (nac_01_mult)
    print ("checking sym:",linalg.norm(nac_01_mult-nac_10_mult))

    print ("Check gradients")
    mc_grad = mc.nuc_grad_method ()
    de_0 = mc_grad.kernel (state=0)
    print (de_0)
    de_1 = mc_grad.kernel (state=1)
    print (de_1)



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver J. Backhouse <olbackhouse@gmail.com>
#         George H. Booth <george.booth@kcl.ac.uk>
#

'''
Auxiliary second-order Green's function perturbation theory for
arbitrary moment consistency
'''

import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__
from pyscf.agf2 import ragf2
from pyscf.agf2 import aux_space as aux


def build_se_part(agf2, eri, gf_occ, gf_vir, os_factor=1.0, ss_factor=1.0):
    ''' Builds either the auxiliaries of the occupied self-energy,
        or virtual if :attr:`gf_occ` and :attr:`gf_vir` are swapped.

    Args:
        eri : _ChemistsERIs
            Electronic repulsion integrals
        gf_occ : GreensFunction
            Occupied Green's function
        gf_vir : GreensFunction
            Virtual Green's function

    Kwargs:
        os_factor : float
            Opposite-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        ss_factor : float
            Same-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0

    Returns:
        :class:`SelfEnergy`
    '''

    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    assert type(gf_occ) is aux.GreensFunction
    assert type(gf_vir) is aux.GreensFunction

    nmo = agf2.nmo
    nocc = gf_occ.naux
    nvir = gf_vir.naux
    naux = nocc * nocc * nvir
    tol = agf2.weight_tol

    if not (agf2.frozen is None or agf2.frozen == 0):
        mask = ragf2.get_frozen_mask(agf2)
        nmo -= np.sum(~mask)

    e = np.zeros((naux))
    v = np.zeros((nmo, naux))

    fpos = np.sqrt(0.5 * os_factor)
    fneg = np.sqrt(0.5 * os_factor + ss_factor)
    fdia = np.sqrt(os_factor)

    eja = lib.direct_sum('j,a->ja', gf_occ.energy, -gf_vir.energy)

    coeffs = (gf_occ.coupling, gf_occ.coupling, gf_vir.coupling)
    qeri = _make_qmo_eris_incore(agf2, eri, coeffs)

    p1 = 0
    for i in range(nocc):
        xija = qeri[:,i,:i].reshape(nmo, -1)
        xjia = qeri[:,:i,i].reshape(nmo, -1)
        xiia = qeri[:,i,i].reshape(nmo, -1)
        eija = gf_occ.energy[i] + eja[:i+1]

        p0, p1 = p1, p1 + i*nvir
        e[p0:p1] = eija[:i].ravel()
        v[:,p0:p1] = fneg * (xija - xjia)

        p0, p1 = p1, p1 + i*nvir
        e[p0:p1] = eija[:i].ravel()
        v[:,p0:p1] = fpos * (xija + xjia)

        p0, p1 = p1, p1 + nvir
        e[p0:p1] = eija[i].ravel()
        v[:,p0:p1] = fdia * xiia

    se = aux.SelfEnergy(e, v, chempot=gf_occ.chempot)
    se.remove_uncoupled(tol=tol)

    if not (agf2.frozen is None or agf2.frozen == 0):
        coupling = np.zeros((agf2.nmo, se.naux))
        coupling[mask] = se.coupling
        se = aux.SelfEnergy(se.energy, coupling, chempot=se.chempot)

    log.timer('se part', *cput0)

    return se


class RAGF2(ragf2.RAGF2):
    ''' Restricted AGF2 with canonical HF reference for arbitrary
        moment consistency

    Attributes:
        nmom : tuple of int
            Compression level of the Green's function and
            self-energy, respectively
        verbose : int
            Print level. Default value equals to :class:`Mole.verbose`
        max_memory : float or int
            Allowed memory in MB. Default value equals to :class:`Mole.max_memory`
        conv_tol : float
            Convergence threshold for AGF2 energy. Default value is 1e-7
        conv_tol_rdm1 : float
            Convergence threshold for first-order reduced density matrix.
            Default value is 1e-8.
        conv_tol_nelec : float
            Convergence threshold for the number of electrons. Default
            value is 1e-6.
        max_cycle : int
            Maximum number of AGF2 iterations. Default value is 50.
        max_cycle_outer : int
            Maximum number of outer Fock loop iterations. Default
            value is 20.
        max_cycle_inner : int
            Maximum number of inner Fock loop iterations. Default
            value is 50.
        weight_tol : float
            Threshold in spectral weight of auxiliaries to be considered
            zero. Default 1e-11.
        fock_diis_space : int
            DIIS space size for Fock loop iterations. Default value is 6.
        fock_diis_min_space :
            Minimum space of DIIS. Default value is 1.
        os_factor : float
            Opposite-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        ss_factor : float
            Same-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        damping : float
            Damping factor for the self-energy. Default value is 0.0

    Saved results

        e_corr : float
            AGF2 correlation energy
        e_tot : float
            Total energy (HF + correlation)
        e_1b : float
            One-body part of :attr:`e_tot`
        e_2b : float
            Two-body part of :attr:`e_tot`
        e_init : float
            Initial correlation energy (truncated MP2)
        converged : bool
            Whether convergence was successful
        se : SelfEnergy
            Auxiliaries of the self-energy
        gf : GreensFunction
            Auxiliaries of the Green's function
    '''

    _keys = {'nmom'}

    def __init__(self, mf, nmom=(None,0), frozen=None, mo_energy=None, mo_coeff=None, mo_occ=None):

        ragf2.RAGF2.__init__(self, mf, frozen=frozen, mo_energy=mo_energy,
                             mo_coeff=mo_coeff, mo_occ=mo_occ)

        self.nmom = nmom

    build_se_part = build_se_part

    def build_se(self, eri=None, gf=None, os_factor=None, ss_factor=None, se_prev=None):
        ''' Builds the auxiliaries of the self-energy.

        Args:
            eri : _ChemistsERIs
                Electronic repulsion integrals
            gf : GreensFunction
                Auxiliaries of the Green's function

        Kwargs:
            os_factor : float
                Opposite-spin factor for spin-component-scaled (SCS)
                calculations. Default 1.0
            ss_factor : float
                Same-spin factor for spin-component-scaled (SCS)
                calculations. Default 1.0
            se_prev : SelfEnergy
                Previous self-energy for damping. Default value is None

        Returns:
            :class:`SelfEnergy`
        '''

        if eri is None: eri = self.ao2mo()
        if gf is None: gf = self.gf
        if gf is None: gf = self.init_gf()

        fock = None
        if self.nmom[0] is not None:
            fock = self.get_fock(eri=eri, gf=gf)

        if os_factor is None: os_factor = self.os_factor
        if ss_factor is None: ss_factor = self.ss_factor

        facs = {'os_factor': os_factor, 'ss_factor': ss_factor}
        gf_occ = gf.get_occupied()
        gf_vir = gf.get_virtual()

        se_occ = self.build_se_part(eri, gf_occ, gf_vir, **facs)
        se_occ = se_occ.compress(n=(None, self.nmom[1]))

        se_vir = self.build_se_part(eri, gf_vir, gf_occ, **facs)
        se_vir = se_vir.compress(n=(None, self.nmom[1]))

        se = aux.combine(se_vir, se_occ)
        se = se.compress(phys=fock, n=(self.nmom[0], None))

        if se_prev is not None and self.damping != 0.0:
            se.coupling *= np.sqrt(1.0-self.damping)
            se_prev.coupling *= np.sqrt(self.damping)
            se = aux.combine(se, se_prev)
            se = se.compress(n=self.nmom)

        return se

    def dump_flags(self, verbose=None):
        ragf2.RAGF2.dump_flags(self, verbose=verbose)
        logger.info(self, 'nmom = %s', repr(self.nmom))
        return self

    def run_diis(self, se, diis=None):
        return se


class _ChemistsERIs(ragf2._ChemistsERIs):
    pass

_make_qmo_eris_incore = ragf2._make_qmo_eris_incore



if __name__ == '__main__':
    from pyscf import gto, scf, mp

    mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='cc-pvdz', verbose=3)
    rhf = scf.RHF(mol)
    rhf.conv_tol = 1e-11
    rhf.run()

    agf2 = RAGF2(rhf, nmom=(None,0))
    agf2.run()

    agf2 = ragf2.RAGF2(rhf)
    agf2.run()



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
G0W0 approximation
'''

#from pyscf.gw.gw import GW
from pyscf.gw import gw_ac
from pyscf.gw import ugw_ac
from pyscf.gw import gw_cd
from pyscf.gw import gw_exact
from pyscf import scf

def GW(mf, freq_int='ac', frozen=None, tdmf=None):
    if isinstance(mf, scf.ghf.GHF):
        raise RuntimeError('GW does not support GHF or GKS methods.')
    if freq_int.lower() == 'ac':
        if isinstance(mf, scf.uhf.UHF):
            return ugw_ac.GWAC(mf, frozen)
        else:
            return gw_ac.GWAC(mf, frozen)
    elif freq_int.lower() == 'cd':
        if isinstance(mf, scf.uhf.UHF):
            raise RuntimeError('GWCD does not support UHF or UKS methods.')
        else:
            return gw_cd.GWCD(mf, frozen)
    elif freq_int.lower() == 'exact':
        if isinstance(mf, scf.uhf.UHF):
            raise RuntimeError('GWExact does not support UHF or UKS methods.')
        else:
            return gw_exact.GWExact(mf, frozen, tdmf)
    else:
        raise RuntimeError("GW frequency integration method %s not recognized. "
                           "Options are 'ac', 'cd', and 'exact'."%(freq_int))



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Lahh dee dah
'''
Multi-configuration pair-density functional theory
==================================================

Simple usage::

    >>> from pyscf import gto, scf, mcpdft
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='def2-tzvp')
    >>> mf = scf.RHF(mol).run ()
    >>> mc = mcpdft.CASSCF (mf, 'tPBE', 6, 6)
    >>> mc.run()
'''


import copy
from pyscf.mcpdft.mcpdft import get_mcpdft_child_class
from pyscf.mcpdft.otfnal import make_hybrid_fnal as hyb
from pyscf import mcscf, gto
from pyscf.lib import logger
from pyscf.mcscf import mc1step, casci

def _sanity_check_of_mol(mc_or_mf_or_mol):
    '''
    Sanity check to ensure input is a mol object, not a cell object.
    Raises an error for cell objects.
    '''
    from pyscf.pbc import gto as pbcgto
    if isinstance(mc_or_mf_or_mol, (mc1step.CASSCF, casci.CASCI)):
        mol = mc_or_mf_or_mol._scf.mol
    elif isinstance(mc_or_mf_or_mol, gto.Mole):
        mol = mc_or_mf_or_mol
    else:
        mol = mc_or_mf_or_mol.mol

    if isinstance(mol, pbcgto.cell.Cell):
        raise NotImplementedError("MCPDFT not implemented for PBC")

# NOTE: As of 02/06/2022, initializing PySCF mcscf classes with a symmetry-enabled molecule
# doesn't work.

def _MCPDFT (mc_class, mc_or_mf_or_mol, ot, ncas, nelecas, ncore=None, frozen=None,
             **kwargs):
    # Raise an error if mol is a cell object.
    _sanity_check_of_mol(mc_or_mf_or_mol)
    if isinstance (mc_or_mf_or_mol, (mc1step.CASSCF, casci.CASCI)):
        mc0 = mc_or_mf_or_mol
        mf_or_mol = mc_or_mf_or_mol._scf
    else:
        mc0 = None
        mf_or_mol = mc_or_mf_or_mol
    if isinstance (mf_or_mol, gto.Mole) and mf_or_mol.symmetry:
        logger.warn (mf_or_mol,
                     'Initializing MC-SCF with a symmetry-adapted Mole object may not work!')
    if frozen is not None: mc1 = mc_class (mf_or_mol, ncas, nelecas, ncore=ncore, frozen=frozen)
    else: mc1 = mc_class (mf_or_mol, ncas, nelecas, ncore=ncore)
    mc2 = get_mcpdft_child_class (mc1, ot, **kwargs)
    if mc0 is not None:
        mc2.verbose = mc0.verbose
        mc2.stdout = mc0.stdout
        mc2.mo_coeff = mc_or_mf_or_mol.mo_coeff.copy ()    
        mc2.ci = copy.deepcopy (mc_or_mf_or_mol.ci)
        mc2.converged = mc0.converged
    return mc2

def CASSCFPDFT (mc_or_mf_or_mol, ot, ncas, nelecas, ncore=None, frozen=None,
                **kwargs):
    return _MCPDFT (mcscf.CASSCF, mc_or_mf_or_mol, ot, ncas, nelecas, ncore=ncore, frozen=frozen,
                    **kwargs)

def CASCIPDFT (mc_or_mf_or_mol, ot, ncas, nelecas, ncore=None, **kwargs):
    return _MCPDFT (mcscf.CASCI, mc_or_mf_or_mol, ot, ncas, nelecas, ncore=ncore,
                    **kwargs)

CASSCF=CASSCFPDFT
CASCI=CASCIPDFT

class MultiStateMCPDFTSolver :
    pass
    # tag


# Monkeypatch for double grad folders
# TODO: more elegant solution
import os
mypath = os.path.dirname (os.path.dirname (os.path.abspath (__file__)))
mygradpath = os.path.join (mypath, 'grad')
from pyscf import grad
grad.__path__.append (mygradpath)
grad.__path__=list(set(grad.__path__))
mydfgradpath = os.path.join (os.path.join (mypath, 'df'), 'grad')
from pyscf.df import grad as df_grad
df_grad.__path__.append (mydfgradpath)
df_grad.__path__=list(set(df_grad.__path__))
mynacpath = os.path.join(mypath, "nac")
from pyscf import nac
nac.__path__.append(mynacpath)
nac.__path__ = list(set(nac.__path__))



#!/usr/bin/env python
# Copyright 2014-2018,2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
A low level interface to libcint library. It's recommended to use the
Mole.intor method to drive the integral evaluation functions.
'''

import warnings
import ctypes
import numpy
from pyscf import lib

libcgto = lib.load_library('libcgto')

ANG_OF     = 1
NPRIM_OF   = 2
NCTR_OF    = 3
KAPPA_OF   = 4
PTR_EXP    = 5
PTR_COEFF  = 6
BAS_SLOTS  = 8
NGRIDS     = 11
PTR_GRIDS  = 12

def getints(intor_name, atm, bas, env, shls_slice=None, comp=None, hermi=0,
            aosym='s1', ao_loc=None, cintopt=None, out=None):
    r'''1e and 2e integral generator.

    Args:
        intor_name : str

            ================================  =============
            Function                          Expression
            ================================  =============
            "int1e_ovlp"                      ( \| \)
            "int1e_nuc"                       ( \| nuc \| \)
            "int1e_kin"                       (.5 \| p dot p\)
            "int1e_ia01p"                     (#C(0 1) \| nabla-rinv \| cross p\)
            "int1e_giao_irjxp"                (#C(0 1) \| r cross p\)
            "int1e_cg_irxp"                   (#C(0 1) \| rc cross p\)
            "int1e_giao_a11part"              (-.5 \| nabla-rinv \| r\)
            "int1e_cg_a11part"                (-.5 \| nabla-rinv \| rc\)
            "int1e_a01gp"                     (g \| nabla-rinv cross p \|\)
            "int1e_igkin"                     (#C(0 .5) g \| p dot p\)
            "int1e_igovlp"                    (#C(0 1) g \|\)
            "int1e_ignuc"                     (#C(0 1) g \| nuc \|\)
            "int1e_z"                         ( \| zc \| \)
            "int1e_zz"                        ( \| zc zc \| \)
            "int1e_r"                         ( \| rc \| \)
            "int1e_r2"                        ( \| rc dot rc \| \)
            "int1e_rr"                        ( \| rc rc \| \)
            "int1e_rrr"                       ( \| rc rc rc \| \)
            "int1e_rrrr"                      ( \| rc rc rc rc \| \)
            "int1e_pnucp"                     (p* \| nuc dot p \| \)
            "int1e_prinvxp"                   (p* \| rinv cross p \| \)
            "int1e_ipovlp"                    (nabla \|\)
            "int1e_ipkin"                     (.5 nabla \| p dot p\)
            "int1e_ipnuc"                     (nabla \| nuc \|\)
            "int1e_iprinv"                    (nabla \| rinv \|\)
            "int1e_rinv"                      (\| rinv \|\)
            "int1e_pnucxp"                    (p* \| nuc cross p \| \)
            "int1e_irp"                       ( \| rc nabla \| \)
            "int1e_irrp"                      ( \| rc rc nabla \| \)
            "int1e_irpr"                      ( \| rc nabla rc \| \)
            "int1e_ggovlp"                    ( \| g g \| \)
            "int1e_ggkin"                     (.5 \| g g p dot p \| \)
            "int1e_ggnuc"                     ( \| g g nuc \| \)
            "int1e_grjxp"                     ( \| g r cross p \| \)
            "ECPscalar"                       AREP ECP integrals, similar to int1e_nuc
            "ECPscalar_ipnuc"                 (nabla i | ECP | ), similar to int1e_ipnuc
            "ECPscalar_iprinv"                similar to int1e_iprinv for a specific atom
            "ECPscalar_ignuc"                 similar to int1e_ignuc
            "ECPscalar_iprinvip"              similar to int1e_iprinvip
            "ECPso"                           < | Spin-orbit ECP | >
            "int1e_ovlp_spinor"               ( \| \)
            "int1e_nuc_spinor"                ( \| nuc \|\)
            "int1e_srsr_spinor"               (sigma dot r \| sigma dot r\)
            "int1e_sr_spinor"                 (sigma dot r \|\)
            "int1e_srsp_spinor"               (sigma dot r \| sigma dot p\)
            "int1e_spsp_spinor"               (sigma dot p \| sigma dot p\)
            "int1e_sp_spinor"                 (sigma dot p \|\)
            "int1e_spnucsp_spinor"            (sigma dot p \| nuc \| sigma dot p\)
            "int1e_srnucsr_spinor"            (sigma dot r \| nuc \| sigma dot r\)
            "int1e_govlp_spinor"              (g \|\)
            "int1e_gnuc_spinor"               (g \| nuc \|\)
            "int1e_cg_sa10sa01_spinor"        (.5 sigma cross rc \| sigma cross nabla-rinv \|\)
            "int1e_cg_sa10sp_spinor"          (.5 rc cross sigma \| sigma dot p\)
            "int1e_cg_sa10nucsp_spinor"       (.5 rc cross sigma \| nuc \| sigma dot p\)
            "int1e_giao_sa10sa01_spinor"      (.5 sigma cross r \| sigma cross nabla-rinv \|\)
            "int1e_giao_sa10sp_spinor"        (.5 r cross sigma \| sigma dot p\)
            "int1e_giao_sa10nucsp_spinor"     (.5 r cross sigma \| nuc \| sigma dot p\)
            "int1e_sa01sp_spinor"             (\| nabla-rinv cross sigma \| sigma dot p\)
            "int1e_spgsp_spinor"              (g sigma dot p \| sigma dot p\)
            "int1e_spgnucsp_spinor"           (g sigma dot p \| nuc \| sigma dot p\)
            "int1e_spgsa01_spinor"            (g sigma dot p \| nabla-rinv cross sigma \|\)
            "int1e_spspsp_spinor"             (sigma dot p \| sigma dot p sigma dot p\)
            "int1e_spnuc_spinor"              (sigma dot p \| nuc \|\)
            "int1e_ipovlp_spinor"             (nabla \|\)
            "int1e_ipkin_spinor"              (.5 nabla \| p dot p\)
            "int1e_ipnuc_spinor"              (nabla \| nuc \|\)
            "int1e_iprinv_spinor"             (nabla \| rinv \|\)
            "int1e_ipspnucsp_spinor"          (nabla sigma dot p \| nuc \| sigma dot p\)
            "int1e_ipsprinvsp_spinor"         (nabla sigma dot p \| rinv \| sigma dot p\)
            "int1e_grids"                     ( \| 1/r_grids \| \)
            "int1e_grids_spinor"              ( \| 1/r_grids \| \)
            "int1e_grids_ip"                  (nabla \| 1/r_grids \| \)
            "int1e_grids_ip_spinor"           (nabla \| 1/r_grids \| \)
            "int1e_grids_spvsp_spinor"        (sigma dot p \| 1/r_grids \| sigma dot p\)
            "int2e"                           ( \, \| \, \)
            "int2e_ig1"                       (#C(0 1) g \, \| \, \)
            "int2e_gg1"                       (g g \, \| \, \)
            "int2e_g1g2"                      (g \, \| g \, \)
            "int2e_p1vxp1"                    ( p* \, cross p \| \, \) ; SSO
            "int2e_spinor"                    (, \| \, \)
            "int2e_spsp1_spinor"              (sigma dot p \, sigma dot p \| \, \)
            "int2e_spsp1spsp2_spinor"         (sigma dot p \, sigma dot p \| sigma dot p \, sigma dot p \)
            "int2e_srsr1_spinor"              (sigma dot r \, sigma dot r \| \,\)
            "int2e_srsr1srsr2_spinor"         (sigma dot r \, sigma dot r \| sigma dot r \, sigma dot r\)
            "int2e_cg_sa10sp1_spinor"         (.5 rc cross sigma \, sigma dot p \| \,\)
            "int2e_cg_sa10sp1spsp2_spinor"    (.5 rc cross sigma \, sigma dot p \| sigma dot p \, sigma dot p \)
            "int2e_giao_sa10sp1_spinor"       (.5 r cross sigma \, sigma dot p \| \,\)
            "int2e_giao_sa10sp1spsp2_spinor"  (.5 r cross sigma \, sigma dot p \| sigma dot p \, sigma dot p \)
            "int2e_g1_spinor"                 (g \, \| \,\)
            "int2e_spgsp1_spinor"             (g sigma dot p \, sigma dot p \| \,\)
            "int2e_g1spsp2_spinor"            (g \, \| sigma dot p \, sigma dot p\)
            "int2e_spgsp1spsp2_spinor"        (g sigma dot p \, sigma dot p \| sigma dot p \, sigma dot p\)
            "int2e_spv1_spinor"               (sigma dot p \, \| \,\)
            "int2e_vsp1_spinor"               (\, sigma dot p \| \,\)
            "int2e_spsp2_spinor"              (\, \| sigma dot p \, sigma dot p\)
            "int2e_spv1spv2_spinor"           (sigma dot p \, \| sigma dot p \,\)
            "int2e_vsp1spv2_spinor"           (\, sigma dot p \| sigma dot p \,\)
            "int2e_spv1vsp2_spinor"           (sigma dot p \, \| \, sigma dot p\)
            "int2e_vsp1vsp2_spinor"           (\, sigma dot p \| \, sigma dot p\)
            "int2e_spv1spsp2_spinor"          (sigma dot p \, \| sigma dot p \, sigma dot p\)
            "int2e_vsp1spsp2_spinor"          (\, sigma dot p \| sigma dot p \, sigma dot p\)
            "int2e_ig1"                       (#C(0 1) g \, \| \, \)
            "int2e_ip1"                       (nabla \, \| \,\)
            "int2e_ip1_spinor"                (nabla \, \| \,\)
            "int2e_ipspsp1_spinor"            (nabla sigma dot p \, sigma dot p \| \,\)
            "int2e_ip1spsp2_spinor"           (nabla \, \| sigma dot p \, sigma dot p\)
            "int2e_ipspsp1spsp2_spinor"       (nabla sigma dot p \, sigma dot p \| sigma dot p \, sigma dot p\)
            "int2e_ipsrsr1_spinor"            (nabla sigma dot r \, sigma dot r \| \,\)
            "int2e_ip1srsr2_spinor"           (nabla \, \| sigma dot r \, sigma dot r\)
            "int2e_ipsrsr1srsr2_spinor"       (nabla sigma dot r \, sigma dot r \| sigma dot r \, sigma dot r\)
            "int2e_ip1"                       (nabla \, \| \,\)
            "int2e_ssp1ssp2_spinor"           ( \, sigma dot p \| gaunt \| \, sigma dot p\)
            "int2e_cg_ssa10ssp2_spinor"       (rc cross sigma \, \| gaunt \| \, sigma dot p\)
            "int2e_giao_ssa10ssp2_spinor"     (r cross sigma  \, \| gaunt \| \, sigma dot p\)
            "int2e_gssp1ssp2_spinor"          (g \, sigma dot p  \| gaunt \| \, sigma dot p\)
            "int2e_ipip1"                     ( nabla nabla \, \| \, \)
            "int2e_ipvip1"                    ( nabla \, nabla \| \, \)
            "int2e_ip1ip2"                    ( nabla \, \| nabla \, \)
            "int3c2e_ip1"                     (nabla \, \| \)
            "int3c2e_ip2"                     ( \, \| nabla\)
            "int2c2e_ip1"                     (nabla \| r12 \| \)
            "int3c2e_spinor"                  (nabla \, \| \)
            "int3c2e_spsp1_spinor"            (nabla \, \| \)
            "int3c2e_ip1_spinor"              (nabla \, \| \)
            "int3c2e_ip2_spinor"              ( \, \| nabla\)
            "int3c2e_ipspsp1_spinor"          (nabla sigma dot p \, sigma dot p \| \)
            "int3c2e_spsp1ip2_spinor"         (sigma dot p \, sigma dot p \| nabla \)
            "ECPscalar_spinor"                AREP ECP integrals, similar to int1e_nuc
            "ECPscalar_ipnuc_spinor"          (nabla i | ECP | ), similar to int1e_ipnuc
            "ECPscalar_iprinv_spinor"         similar to int1e_iprinv for a specific atom
            "ECPscalar_ignuc_spinor"          similar to int1e_ignuc
            "ECPscalar_iprinvip_spinor"       similar to int1e_iprinvip
            "ECPso_spinor"                    < | sigam dot Spin-orbit ECP | >
            ================================  =============

            ==========   =============
            Symbol       Meaning
            ==========   =============
            p            :math:`-i \nabla`
            ip           :math:`\nabla`
            r0           :math:`\vec{r} - (0,0,0)`
            rc           :math:`\vec{r} - \vec{R}_{(\texttt{env[PTR_COMMON_ORIG]})}`
            ri           :math:`\vec{r} - \vec{R}_i`
            rj           :math:`\vec{r} - \vec{R}_j`
            rk           :math:`\vec{r} - \vec{R}_k`
            rl           :math:`\vec{r} - \vec{R}_l`
            r            can be ri/rj/rk/rl; associate with the basis it operates
            g            :math:`(i/2) (\vec{R}_{bra} - \vec{R}_{ket}) \times \vec{r}`
            sigma        three pauli matrices
            dot, cross   can be used to combine operator-bra or operator-ket
            rinv         :math:`|\vec{r} - \vec{R}_{(\texttt{env[PTR_RINV_ORIG]})}|^{-1}`
            nuc          :math:`\sum_N Z_N |\vec{r} - \vec{R}_N|^{-1}`
            nabla-rinv   :math:`\nabla |\vec{r} - \vec{R}_{(\texttt{env[PTR_RINV_ORIG]})}|^{-1}`
            gaunt        :math:`\alpha_i \cdot \alpha_j  / |\vec{r}_i - \vec{r}_j|`, \
                         note the minus sign in Gaunt is not included
            breit        :math:`-(\alpha_i \cdot \alpha_j)/(2 |\vec{r}_i - \vec{r}_j|) \
                         - ((\alpha_i \cdot r_{ij})  (\alpha_j \cdot r_{ij})) / \
                         (2|\vec{r}_i - \vec{r}_j|^3)`
            ==========   =============

        atm : int32 ndarray
            libcint integral function argument
        bas : int32 ndarray
            libcint integral function argument
        env : float64 ndarray
            libcint integral function argument

    Kwargs:
        shls_slice : 8-element list
            (ish_start, ish_end, jsh_start, jsh_end, ksh_start, ksh_end, lsh_start, lsh_end)
        comp : int
            Components of the integrals, e.g. int1e_ipovlp has 3 components.
        hermi : int (1e integral only)
            Symmetry of the 1e integrals

            | 0 : no symmetry assumed (default)
            | 1 : hermitian
            | 2 : anti-hermitian

        aosym : str (2e integral only)
            Symmetry of the 2e integrals

            | 4 or '4' or 's4': 4-fold symmetry (default)
            | '2ij' or 's2ij' : symmetry between i, j in (ij|kl)
            | '2kl' or 's2kl' : symmetry between k, l in (ij|kl)
            | 1 or '1' or 's1': no symmetry

        out : ndarray (2e integral only)
            array to store the 2e AO integrals

    Returns:
        ndarray of 1-electron integrals, can be either 2-dim or 3-dim, depending on comp

    Examples:

    >>> mol.build(atom='H 0 0 0; H 0 0 1.1', basis='sto-3g')
    >>> gto.getints('int1e_ipnuc_sph', mol._atm, mol._bas, mol._env, comp=3) # <nabla i | V_nuc | j>
    [[[ 0.          0.        ]
      [ 0.          0.        ]]
     [[ 0.          0.        ]
      [ 0.          0.        ]]
     [[ 0.10289944  0.48176097]
      [-0.48176097 -0.10289944]]]
    '''
    intor_name, comp = _get_intor_and_comp(intor_name, comp)
    if any(bas[:,ANG_OF] > 12):
        raise NotImplementedError('cint library does not support high angular (l>12) GTOs')

    if (intor_name.startswith('int1e') or
        intor_name.startswith('ECP') or
        intor_name.startswith('int2c2e')):
        return getints2c(intor_name, atm, bas, env, shls_slice, comp,
                         hermi, ao_loc, cintopt, out)
    elif intor_name.startswith('int2e') or intor_name.startswith('int4c1e'):
        return getints4c(intor_name, atm, bas, env, shls_slice, comp,
                         aosym, ao_loc, cintopt, out)
    elif intor_name.startswith('int3c'):
        return getints3c(intor_name, atm, bas, env, shls_slice, comp,
                         aosym, ao_loc, cintopt, out)
    else:
        raise KeyError('Unknown intor %s' % intor_name)

def _get_intor_and_comp(intor_name, comp=None):
    intor_name = ascint3(intor_name)
    if comp is None:
        try:
            if '_spinor' in intor_name:
                fname = intor_name.replace('_spinor', '')
                comp = _INTOR_FUNCTIONS[fname][1]
            else:
                fname = intor_name.replace('_sph', '').replace('_cart', '')
                comp = _INTOR_FUNCTIONS[fname][0]
        except KeyError:
            warnings.warn('Function %s not found.  Set its comp to 1' % intor_name)
            comp = 1
    return intor_name, comp

_INTOR_FUNCTIONS = {
    # Functiona name            : (comp-for-scalar, comp-for-spinor)
    'int1e_ovlp'                : (1, 1),
    'int1e_nuc'                 : (1, 1),
    'int1e_kin'                 : (1, 1),
    'int1e_ia01p'               : (3, 3),
    'int1e_giao_irjxp'          : (3, 3),
    'int1e_cg_irxp'             : (3, 3),
    'int1e_giao_a11part'        : (9, 9),
    'int1e_cg_a11part'          : (9, 9),
    'int1e_a01gp'               : (9, 9),
    'int1e_igkin'               : (3, 3),
    'int1e_igovlp'              : (3, 3),
    'int1e_ignuc'               : (3, 3),
    'int1e_pnucp'               : (1, 1),
    'int1e_z'                   : (1, 1),
    'int1e_zz'                  : (1, 1),
    'int1e_r'                   : (3, 3),
    'int1e_r2'                  : (1, 1),
    'int1e_r4'                  : (1, 1),
    'int1e_rr'                  : (9, 9),
    'int1e_rrr'                 : (27, 27),
    'int1e_rrrr'                : (81, 81),
    'int1e_z_origj'             : (1, 1),
    'int1e_zz_origj'            : (1, 1),
    'int1e_r_origj'             : (3, 3),
    'int1e_rr_origj'            : (9, 9),
    'int1e_r2_origj'            : (1, 1),
    'int1e_r4_origj'            : (1, 1),
    'int1e_p4'                  : (1, 1),
    'int1e_prinvp'              : (1, 1),
    'int1e_prinvxp'             : (3, 3),
    'int1e_pnucxp'              : (3, 3),
    'int1e_irp'                 : (9, 9),
    'int1e_irrp'                : (27, 27),
    'int1e_irpr'                : (27, 27),
    'int1e_ggovlp'              : (9, 9),
    'int1e_ggkin'               : (9, 9),
    'int1e_ggnuc'               : (9, 9),
    'int1e_grjxp'               : (9, 9),
    'int2e'                     : (1, 1),
    'int2e_ig1'                 : (3, 3),
    'int2e_gg1'                 : (9, 9),
    'int2e_g1g2'                : (9, 9),
    'int2e_ip1v_rc1'            : (9, 9),
    'int2e_ip1v_r1'             : (9, 9),
    'int2e_ipvg1_xp1'           : (9, 9),
    'int2e_ipvg2_xp1'           : (9, 9),
    'int2e_p1vxp1'              : (3, 3),
    'int1e_inuc_rcxp'           : (3, 3),
    'int1e_inuc_rxp'            : (3, 3),
    'int1e_sigma'               : (12,3),
    'int1e_spsigmasp'           : (12,3),
    'int1e_srsr'                : (4, 1),
    'int1e_sr'                  : (4, 1),
    'int1e_srsp'                : (4, 1),
    'int1e_spsp'                : (4, 1),
    'int1e_sp'                  : (4, 1),
    'int1e_spnucsp'             : (4, 1),
    'int1e_sprinvsp'            : (4, 1),
    'int1e_srnucsr'             : (4, 1),
    'int1e_sprsp'               : (12,3),
    'int1e_govlp'               : (3, 3),
    'int1e_gnuc'                : (3, 3),
    'int1e_cg_sa10sa01'         : (36,9),
    'int1e_cg_sa10sp'           : (12,3),
    'int1e_cg_sa10nucsp'        : (12,3),
    'int1e_giao_sa10sa01'       : (36,9),
    'int1e_giao_sa10sp'         : (12,3),
    'int1e_giao_sa10nucsp'      : (12,3),
    'int1e_sa01sp'              : (12,3),
    'int1e_spgsp'               : (12,3),
    'int1e_spgnucsp'            : (12,3),
    'int1e_spgsa01'             : (36,9),
    'int2e_spsp1'               : (4, 1),
    'int2e_spsp1spsp2'          : (16,1),
    'int2e_srsr1'               : (4, 1),
    'int2e_srsr1srsr2'          : (16,1),
    'int2e_cg_sa10sp1'          : (12,3),
    'int2e_cg_sa10sp1spsp2'     : (48,3),
    'int2e_giao_sa10sp1'        : (12,3),
    'int2e_giao_sa10sp1spsp2'   : (48,3),
    'int2e_g1'                  : (12,3),
    'int2e_spgsp1'              : (12,3),
    'int2e_g1spsp2'             : (12,3),
    'int2e_spgsp1spsp2'         : (48,3),
    'int2e_pp1'                 : (1, 1),
    'int2e_pp2'                 : (1, 1),
    'int2e_pp1pp2'              : (1, 1),
    'int1e_spspsp'              : (4, 1),
    'int1e_spnuc'               : (4, 1),
    'int2e_spv1'                : (4, 1),
    'int2e_vsp1'                : (4, 1),
    'int2e_spsp2'               : (4, 1),
    'int2e_spv1spv2'            : (16,1),
    'int2e_vsp1spv2'            : (16,1),
    'int2e_spv1vsp2'            : (16,1),
    'int2e_vsp1vsp2'            : (16,1),
    'int2e_spv1spsp2'           : (16,1),
    'int2e_vsp1spsp2'           : (16,1),
    'int1e_ipovlp'              : (3, 3),
    'int1e_ipkin'               : (3, 3),
    'int1e_ipnuc'               : (3, 3),
    'int1e_iprinv'              : (3, 3),
    'int1e_rinv'                : (1, 1),
    'int1e_ipspnucsp'           : (12,3),
    'int1e_ipsprinvsp'          : (12,3),
    'int1e_ippnucp'             : (3, 3),
    'int1e_ipprinvp'            : (3, 3),
    'int2e_ip1'                 : (3, 3),
    'int2e_ip2'                 : (3, 3),
    'int2e_ipspsp1'             : (12,3),
    'int2e_ip1spsp2'            : (12,3),
    'int2e_ipspsp1spsp2'        : (48,3),
    'int2e_ipsrsr1'             : (12,3),
    'int2e_ip1srsr2'            : (12,3),
    'int2e_ipsrsr1srsr2'        : (48,3),
    'int2e_ssp1ssp2'            : (16,1),
    'int2e_ssp1sps2'            : (16,1),
    'int2e_sps1ssp2'            : (16,1),
    'int2e_sps1sps2'            : (16,1),
    'int2e_cg_ssa10ssp2'        : (48,3),
    'int2e_giao_ssa10ssp2'      : (18,3),
    'int2e_gssp1ssp2'           : (18,3),
    'int2e_gauge_r1_ssp1ssp2'   : (None, 1),
    'int2e_gauge_r1_ssp1sps2'   : (None, 1),
    'int2e_gauge_r1_sps1ssp2'   : (None, 1),
    'int2e_gauge_r1_sps1sps2'   : (None, 1),
    'int2e_gauge_r2_ssp1ssp2'   : (None, 1),
    'int2e_gauge_r2_ssp1sps2'   : (None, 1),
    'int2e_gauge_r2_sps1ssp2'   : (None, 1),
    'int2e_gauge_r2_sps1sps2'   : (None, 1),
    'int1e_ipipovlp'            : (9, 9),
    'int1e_ipovlpip'            : (9, 9),
    'int1e_ipipkin'             : (9, 9),
    'int1e_ipkinip'             : (9, 9),
    'int1e_ipipnuc'             : (9, 9),
    'int1e_ipnucip'             : (9, 9),
    'int1e_ipiprinv'            : (9, 9),
    'int1e_iprinvip'            : (9, 9),
    'int2e_ipip1'               : (9, 9),
    'int2e_ipvip1'              : (9, 9),
    'int2e_ip1ip2'              : (9, 9),
    'int1e_ipippnucp'           : (9, 9),
    'int1e_ippnucpip'           : (9, 9),
    'int1e_ipipprinvp'          : (9, 9),
    'int1e_ipprinvpip'          : (9, 9),
    'int1e_ipipspnucsp'         : (36,9),
    'int1e_ipspnucspip'         : (36,9),
    'int1e_ipipsprinvsp'        : (36,9),
    'int1e_ipsprinvspip'        : (36,9),
    'int3c2e'                   : (1, 1),
    'int3c2e_ip1'               : (3, 3),
    'int3c2e_ip2'               : (3, 3),
    'int3c2e_pvp1'              : (1, 1),
    'int3c2e_pvxp1'             : (3, 3),
    'int2c2e_ip1'               : (3, 3),
    'int2c2e_ip2'               : (3, 3),
    'int3c2e_ig1'               : (3, 3),
    'int3c2e_spsp1'             : (4, 1),
    'int3c2e_ipspsp1'           : (12,3),
    'int3c2e_spsp1ip2'          : (12,3),
    'int3c2e_ipip1'             : (9, 9),
    'int3c2e_ipip2'             : (9, 9),
    'int3c2e_ipvip1'            : (9, 9),
    'int3c2e_ip1ip2'            : (9, 9),
    'int2c2e_ip1ip2'            : (9, 9),
    'int2c2e_ipip1'             : (9, 9),
    'int3c1e'                   : (1, 1),
    'int3c1e_ip1'               : (3, 3),
    'int3c1e_p2'                : (1, 1),
    'int3c1e_iprinv'            : (3, 3),
    'int2c2e'                   : (1, 1),
    'int2e_yp'                  : (1, 1),
    'int2e_stg'                 : (1, 1),
    'int2e_coulerf'             : (1, 1),
    "int1e_grids"               : (1, 1),
    "int1e_grids_ip"            : (3, 3),
    "int1e_grids_spvsp"         : (4, 1),
    'ECPscalar'                 : (1, None),
    'ECPscalar_ipnuc'           : (3, None),
    'ECPscalar_iprinv'          : (3, None),
    'ECPscalar_ignuc'           : (3, None),
    'ECPscalar_iprinvip'        : (9, None),
    'ECPso'                     : (3, 1),
}

def getints2c(intor_name, atm, bas, env, shls_slice=None, comp=1, hermi=0,
              ao_loc=None, cintopt=None, out=None):
    atm = numpy.asarray(atm, dtype=numpy.int32, order='C')
    bas = numpy.asarray(bas, dtype=numpy.int32, order='C')
    env = numpy.asarray(env, dtype=numpy.double, order='C')
    natm = atm.shape[0]
    nbas = bas.shape[0]
    if shls_slice is None:
        shls_slice = (0, nbas, 0, nbas)
    else:
        assert (shls_slice[1] <= nbas and shls_slice[3] <= nbas)
    if ao_loc is None:
        ao_loc = make_loc(bas, intor_name)

    i0, i1, j0, j1 = shls_slice[:4]
    naoi = ao_loc[i1] - ao_loc[i0]
    naoj = ao_loc[j1] - ao_loc[j0]

    if 'int1e_grids' in intor_name:
        shape = (int(env[NGRIDS]), naoi, naoj, comp)
        prefix = 'GTOgrids_'
    else:
        shape = (naoi, naoj, comp)
        prefix = 'GTO'

    if intor_name.endswith('_cart') or intor_name.endswith('_sph'):
        dtype = numpy.double
        drv_name = prefix + 'int2c'
    else:
        dtype = numpy.complex128
        if '2c2e' in intor_name:
            assert (hermi != lib.HERMITIAN and
                   hermi != lib.ANTIHERMI)
        drv_name = prefix + 'int2c_spinor'

    mat = numpy.ndarray(shape, dtype, out, order='F')

    if mat.size > 0:
        if cintopt is None:
            cintopt = make_cintopt(atm, bas, env, intor_name)

        fn = getattr(libcgto, drv_name)
        fn(getattr(libcgto, intor_name), mat.ctypes.data_as(ctypes.c_void_p),
           ctypes.c_int(comp), ctypes.c_int(hermi),
           (ctypes.c_int*4)(*(shls_slice[:4])),
           ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
           atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(natm),
           bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(nbas),
           env.ctypes.data_as(ctypes.c_void_p))

    mat = numpy.rollaxis(mat, -1, 0)
    if comp == 1:
        mat = mat[0]
    return mat

def getints3c(intor_name, atm, bas, env, shls_slice=None, comp=1,
              aosym='s1', ao_loc=None, cintopt=None, out=None):
    atm = numpy.asarray(atm, dtype=numpy.int32, order='C')
    bas = numpy.asarray(bas, dtype=numpy.int32, order='C')
    env = numpy.asarray(env, dtype=numpy.double, order='C')
    natm = atm.shape[0]
    nbas = bas.shape[0]
    if shls_slice is None:
        shls_slice = (0, nbas, 0, nbas, 0, nbas)
        if 'ssc' in intor_name or 'spinor' in intor_name:
            bas = numpy.asarray(numpy.vstack((bas,bas)), dtype=numpy.int32)
            shls_slice = (0, nbas, 0, nbas, nbas, nbas*2)
            nbas = bas.shape[0]
    else:
        assert (shls_slice[1] <= nbas and
               shls_slice[3] <= nbas and
               shls_slice[5] <= nbas)

    i0, i1, j0, j1, k0, k1 = shls_slice[:6]
    if ao_loc is None:
        ao_loc = make_loc(bas, intor_name)
        if 'ssc' in intor_name:
            ao_loc[k0:] = ao_loc[k0] + make_loc(bas[k0:], 'cart')
        elif 'spinor' in intor_name:
            # The auxbasis for electron-2 is in real spherical representation
            ao_loc[k0:] = ao_loc[k0] + make_loc(bas[k0:], 'sph')

    naok = ao_loc[k1] - ao_loc[k0]

    if aosym in ('s1',):
        naoi = ao_loc[i1] - ao_loc[i0]
        naoj = ao_loc[j1] - ao_loc[j0]
        shape = (naoi, naoj, naok, comp)
    else:
        aosym = 's2ij'
        nij = ao_loc[i1]*(ao_loc[i1]+1)//2 - ao_loc[i0]*(ao_loc[i0]+1)//2
        shape = (nij, naok, comp)

    if 'spinor' in intor_name:
        mat = numpy.ndarray(shape, numpy.complex128, out, order='F')
        drv = libcgto.GTOr3c_drv
        fill = getattr(libcgto, 'GTOr3c_fill_'+aosym)
    else:
        mat = numpy.ndarray(shape, numpy.double, out, order='F')
        drv = libcgto.GTOnr3c_drv
        fill = getattr(libcgto, 'GTOnr3c_fill_'+aosym)

    if mat.size > 0:
        # Generating opt for all indices leads to large overhead and poor OMP
        # speedup for solvent model and COSX functions. In these methods,
        # the third index of the three center integrals corresponds to a
        # large number of grids. Initializing the opt for the third index is
        # not necessary.
        if cintopt is None:
            if '3c2e' in intor_name:
                # int3c2e opt without the 3rd index.
                cintopt = make_cintopt(atm, bas[:max(i1, j1)], env, intor_name)
            else:
                cintopt = make_cintopt(atm, bas, env, intor_name)

        drv(getattr(libcgto, intor_name), fill,
            mat.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(comp),
            (ctypes.c_int*6)(*(shls_slice[:6])),
            ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
            atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(natm),
            bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(nbas),
            env.ctypes.data_as(ctypes.c_void_p))

    mat = numpy.rollaxis(mat, -1, 0)
    if comp == 1:
        mat = mat[0]
    return mat

def getints4c(intor_name, atm, bas, env, shls_slice=None, comp=1,
              aosym='s1', ao_loc=None, cintopt=None, out=None):
    aosym = _stand_sym_code(aosym)
    atm = numpy.asarray(atm, dtype=numpy.int32, order='C')
    bas = numpy.asarray(bas, dtype=numpy.int32, order='C')
    env = numpy.asarray(env, dtype=numpy.double, order='C')
    c_atm = atm.ctypes.data_as(ctypes.c_void_p)
    c_bas = bas.ctypes.data_as(ctypes.c_void_p)
    c_env = env.ctypes.data_as(ctypes.c_void_p)
    natm = atm.shape[0]
    nbas = bas.shape[0]
    ao_loc = make_loc(bas, intor_name)

    if '_spinor' in intor_name:
        assert (aosym == 's1')

    if aosym == 's8':
        assert (shls_slice is None)
        from pyscf.scf import _vhf
        nao = int(ao_loc[-1])
        nao_pair = nao*(nao+1)//2
        out = numpy.ndarray((nao_pair*(nao_pair+1)//2), buffer=out)
        if nao_pair == 0:
            return out

        if cintopt is None:
            cintopt = make_cintopt(atm, bas, env, intor_name)
        drv = _vhf.libcvhf.GTO2e_cart_or_sph
        drv(getattr(libcgto, intor_name), cintopt,
            out.ctypes.data_as(ctypes.c_void_p),
            ao_loc.ctypes.data_as(ctypes.c_void_p),
            c_atm, ctypes.c_int(natm), c_bas, ctypes.c_int(nbas), c_env)
        return out

    else:
        if shls_slice is None:
            shls_slice = (0, nbas, 0, nbas, 0, nbas, 0, nbas)
        elif len(shls_slice) == 4:
            shls_slice = shls_slice + (0, nbas, 0, nbas)
        else:
            assert (shls_slice[1] <= nbas and shls_slice[3] <= nbas and
                   shls_slice[5] <= nbas and shls_slice[7] <= nbas)
        i0, i1, j0, j1, k0, k1, l0, l1 = shls_slice
        naoi = ao_loc[i1] - ao_loc[i0]
        naoj = ao_loc[j1] - ao_loc[j0]
        naok = ao_loc[k1] - ao_loc[k0]
        naol = ao_loc[l1] - ao_loc[l0]
        if aosym in ('s4', 's2ij'):
            nij = [naoi * (naoi + 1) // 2]
            assert (numpy.all(ao_loc[i0:i1]-ao_loc[i0] == ao_loc[j0:j1]-ao_loc[j0]))
        else:
            nij = [naoi, naoj]
        if aosym in ('s4', 's2kl'):
            nkl = [naok * (naok + 1) // 2]
            assert (numpy.all(ao_loc[k0:k1]-ao_loc[k0] == ao_loc[l0:l1]-ao_loc[l0]))
        else:
            nkl = [naok, naol]
        shape = [comp] + nij + nkl

        if '_spinor' in intor_name:
            drv = libcgto.GTOr4c_drv
            fill = libcgto.GTOr4c_fill_s1
            out = numpy.ndarray(shape[::-1], dtype=numpy.complex128, buffer=out, order='F')
            out = numpy.rollaxis(out, -1, 0)
        else:
            drv = libcgto.GTOnr2e_fill_drv
            fill = getattr(libcgto, 'GTOnr2e_fill_'+aosym)
            out = numpy.ndarray(shape, buffer=out)

        if out.size > 0:
            if cintopt is None:
                cintopt = make_cintopt(atm, bas, env, intor_name)
            prescreen = lib.c_null_ptr()
            drv(getattr(libcgto, intor_name), fill, prescreen,
                out.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(comp),
                (ctypes.c_int*8)(*shls_slice),
                ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
                c_atm, ctypes.c_int(natm), c_bas, ctypes.c_int(nbas), c_env)

        if comp == 1:
            out = out[0]
        return out

def getints_by_shell(intor_name, shls, atm, bas, env, comp=1):
    r'''For given 2, 3 or 4 shells, interface for libcint to get 1e, 2e,
    2-center-2e or 3-center-2e integrals

    Args:
        intor_name : str
            See also :func:`getints` for the supported intor_name
        shls : list of int
            The AO shell-ids of the integrals
        atm : int32 ndarray
            libcint integral function argument
        bas : int32 ndarray
            libcint integral function argument
        env : float64 ndarray
            libcint integral function argument

    Kwargs:
        comp : int
            Components of the integrals, e.g. int1e_ipovlp has 3 components.

    Returns:
        ndarray of 2-dim to 5-dim, depending on the integral type (1e,
        2e, 3c-2e, 2c2e) and the value of comp

    Examples:
        The gradients of the spherical 2e integrals

    >>> mol.build(atom='H 0 0 0; H 0 0 1.1', basis='sto-3g')
    >>> gto.getints_by_shell('int2e_ip1_sph', (0,1,0,1), mol._atm, mol._bas, mol._env, comp=3)
    [[[[[-0.        ]]]]
      [[[[-0.        ]]]]
      [[[[-0.08760462]]]]]
    '''
    intor_name, comp = _get_intor_and_comp(intor_name, comp)

    atm = numpy.asarray(atm, dtype=numpy.int32, order='C')
    bas = numpy.asarray(bas, dtype=numpy.int32, order='C')
    env = numpy.asarray(env, dtype=numpy.double, order='C')
    natm = ctypes.c_int(atm.shape[0])
    nbas = ctypes.c_int(bas.shape[0])
    if intor_name.endswith('_cart'):
        dtype = numpy.double
        def num_cgto_of(basid):
            l = bas[basid,ANG_OF]
            return (l+1)*(l+2)//2 * bas[basid,NCTR_OF]
    elif intor_name.endswith('_sph'):
        dtype = numpy.double
        def num_cgto_of(basid):
            l = bas[basid,ANG_OF]
            return (l*2+1) * bas[basid,NCTR_OF]
    else:
        from pyscf.gto.mole import len_spinor
        dtype = numpy.complex128
        def num_cgto_of(basid):
            l = bas[basid,ANG_OF]
            k = bas[basid,KAPPA_OF]
            return len_spinor(l,k) * bas[basid,NCTR_OF]

    null = lib.c_null_ptr()
    if intor_name.startswith('int3c'):
        assert (len(shls) == 3)
        di = num_cgto_of(shls[0])
        dj = num_cgto_of(shls[1])
        l = bas[shls[2],ANG_OF]
        if intor_name.endswith('_ssc'): # mixed spherical-cartesian
            dk = (l+1)*(l+2)//2 * bas[shls[2],NCTR_OF]
        elif intor_name.endswith('_cart'):
            dk = (l+1)*(l+2)//2 * bas[shls[2],NCTR_OF]
        else:
            dk = (l*2+1) * bas[shls[2],NCTR_OF]
        buf = numpy.empty((di,dj,dk,comp), dtype, order='F')
        fintor = getattr(libcgto, intor_name)
        fintor(buf.ctypes.data_as(ctypes.c_void_p),
               null, (ctypes.c_int*3)(*shls),
               atm.ctypes.data_as(ctypes.c_void_p), natm,
               bas.ctypes.data_as(ctypes.c_void_p), nbas,
               env.ctypes.data_as(ctypes.c_void_p), null, null)
        if comp == 1:
            return buf.reshape(di,dj,dk)
        else:
            return buf.transpose(3,0,1,2)

    elif intor_name.startswith('int2e') or intor_name.startswith('int4c'):
        assert (len(shls) == 4)
        di, dj, dk, dl = [num_cgto_of(x) for x in shls]
        buf = numpy.empty((di,dj,dk,dl,comp), dtype, order='F')
        fintor = getattr(libcgto, intor_name)
        fintor(buf.ctypes.data_as(ctypes.c_void_p),
               null, (ctypes.c_int*4)(*shls),
               atm.ctypes.data_as(ctypes.c_void_p), natm,
               bas.ctypes.data_as(ctypes.c_void_p), nbas,
               env.ctypes.data_as(ctypes.c_void_p), null, null)
        if comp == 1:
            return buf.reshape(di,dj,dk,dl)
        else:
            return buf.transpose(4,0,1,2,3)

    elif (intor_name.startswith('int2c') or '1e' in intor_name or
          'ECP' in intor_name):
        assert (len(shls) == 2)
        di = num_cgto_of(shls[0])
        dj = num_cgto_of(shls[1])
        buf = numpy.empty((di,dj,comp), dtype, order='F')
        fintor = getattr(libcgto, intor_name)
        fintor(buf.ctypes.data_as(ctypes.c_void_p),
               null, (ctypes.c_int*2)(*shls),
               atm.ctypes.data_as(ctypes.c_void_p), natm,
               bas.ctypes.data_as(ctypes.c_void_p), nbas,
               env.ctypes.data_as(ctypes.c_void_p), null, null)
        if comp == 1:
            return buf.reshape(di,dj)
        else:
            return buf.transpose(2,0,1)

    else:
        raise RuntimeError('Unknown intor %s' % intor_name)


def make_loc(bas, key):
    if 'cart' in key:
        l = bas[:,ANG_OF]
        dims = (l+1)*(l+2)//2 * bas[:,NCTR_OF]
    elif 'sph' in key:
        dims = (bas[:,ANG_OF]*2+1) * bas[:,NCTR_OF]
    else:  # spinor
        l = bas[:,ANG_OF]
        k = bas[:,KAPPA_OF]
        dims = (l*4+2) * bas[:,NCTR_OF]
        dims[k<0] = (l[k<0] * 2 + 2) * bas[k<0,NCTR_OF]
        dims[k>0] = (l[k>0] * 2    ) * bas[k>0,NCTR_OF]

    ao_loc = numpy.empty(len(dims)+1, dtype=numpy.int32)
    ao_loc[0] = 0
    dims.cumsum(dtype=numpy.int32, out=ao_loc[1:])
    return ao_loc

def make_cintopt(atm, bas, env, intor):
    intor = intor.replace('_sph','').replace('_cart','').replace('_spinor','')
    c_atm = numpy.asarray(atm, dtype=numpy.int32, order='C')
    c_bas = numpy.asarray(bas, dtype=numpy.int32, order='C')
    c_env = numpy.asarray(env, dtype=numpy.double, order='C')
    natm = c_atm.shape[0]
    nbas = c_bas.shape[0]
    cintopt = ctypes.cast(lib.c_null_ptr(), _cintoptHandler)
    cintopt.intor = intor

    # TODO: call specific ECP optimizers for each intor.
    if intor[:3] == 'ECP':
        AS_ECPBAS_OFFSET = 18  # from gto/mole.py
        if env[AS_ECPBAS_OFFSET] == 0:
            raise RuntimeError('ecpbas or env is not properly initialized')
        foptinit = libcgto.ECPscalar_optimizer
    else:
        foptinit = getattr(libcgto, intor+'_optimizer')
    foptinit(ctypes.byref(cintopt),
             c_atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(natm),
             c_bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(nbas),
             c_env.ctypes.data_as(ctypes.c_void_p))
    return cintopt

class _cintoptHandler(ctypes.c_void_p):
    def __del__(self):
        try:
            if self.intor[:3] == 'ECP':
                libcgto.ECPdel_optimizer(ctypes.byref(self))
            else:
                libcgto.CINTdel_optimizer(ctypes.byref(self))
        except AttributeError:
            pass

def _stand_sym_code(sym):
    if isinstance(sym, int):
        return 's%d' % sym
    elif sym[0] in 'sS':
        return sym.lower()
    else:
        return 's' + sym.lower()

def ascint3(intor_name):
    '''convert cint2 function name to cint3 function name'''
    if intor_name.startswith('cint'):
        intor_name = intor_name[1:]
    if not intor_name.endswith(('_sph', '_cart', '_spinor', '_ssc')):
        intor_name = intor_name + '_spinor'
    return intor_name


if __name__ == '__main__':
    from pyscf import gto
    mol = gto.Mole()
    mol.verbose = 0
    mol.output = None

    mol.atom.extend([
        ["H", (0,  0, 0  )],
        ["H", (0,  0, 1  )],
    ])
    mol.basis = {"H": 'cc-pvdz'}
    mol.build()
    mol.set_rinv_origin(mol.atom_coord(0))
    for i in range(mol.nbas):
        for j in range(mol.nbas):
            print(i, j, getints_by_shell('int1e_prinvxp_sph', (i,j),
                                         mol._atm, mol._bas, mol._env, 3))



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyscf import scf
from pyscf.ci import cisd
from pyscf.ci import ucisd
from pyscf.ci import gcisd
from pyscf.cc import qcisd

def CISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    if mf.istype('UHF'):
        return UCISD(mf, frozen, mo_coeff, mo_occ)
    elif mf.istype('ROHF'):
        from pyscf import lib
        lib.logger.warn(mf, 'RCISD method does not support ROHF method. ROHF object '
                        'is converted to UHF object and UCISD method is called.')
        return UCISD(mf, frozen, mo_coeff, mo_occ)
    else:
        return RCISD(mf, frozen, mo_coeff, mo_occ)
CISD.__doc__ = cisd.CISD.__doc__

def RCISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.df.df_jk import _DFHF

    mf = mf.remove_soscf()
    if not mf.istype('RHF'):
        mf = mf.to_rhf()

    if isinstance(mf, _DFHF) and mf.with_df:
        from pyscf import lib
        lib.logger.warn(mf, f'DF-RCISD for DFHF method {mf} is not available. '
                        'Normal RCISD method is called.')
        return cisd.RCISD(mf, frozen, mo_coeff, mo_occ)
    else:
        return cisd.RCISD(mf, frozen, mo_coeff, mo_occ)
RCISD.__doc__ = cisd.RCISD.__doc__

def UCISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.df.df_jk import _DFHF

    mf = mf.remove_soscf()
    if not mf.istype('UHF'):
        mf = mf.to_uhf()

    if isinstance(mf, _DFHF) and mf.with_df:
        from pyscf import lib
        lib.logger.warn(mf, f'DF-UCISD for DFHF method {mf} is not available. '
                        'Normal UCISD method is called.')
        return ucisd.UCISD(mf, frozen, mo_coeff, mo_occ)
    else:
        return ucisd.UCISD(mf, frozen, mo_coeff, mo_occ)
UCISD.__doc__ = ucisd.UCISD.__doc__


def GCISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.df.df_jk import _DFHF

    mf = mf.remove_soscf()
    if not mf.istype('GHF'):
        mf = mf.to_ghf()

    if isinstance(mf, _DFHF) and mf.with_df:
        raise NotImplementedError('DF-GCISD')
    else:
        return gcisd.GCISD(mf, frozen, mo_coeff, mo_occ)
GCISD.__doc__ = gcisd.GCISD.__doc__


def QCISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    if mf.istype('UHF'):
        raise NotImplementedError
    elif mf.istype('GHF'):
        raise NotImplementedError
    else:
        return RQCISD(mf, frozen, mo_coeff, mo_occ)
QCISD.__doc__ = qcisd.QCISD.__doc__

scf.hf.SCF.QCISD = QCISD

def RQCISD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    import numpy
    from pyscf import lib

    if mf.istype('UHF'):
        raise RuntimeError('RQCISD cannot be used with UHF method.')
    elif mf.istype('ROHF'):
        lib.logger.warn(mf, 'RQCISD method does not support ROHF method. ROHF object '
                        'is converted to UHF object and UQCISD method is called.')
        raise NotImplementedError

    mf = mf.remove_soscf()
    if not mf.istype('RHF'):
        mf = mf.to_rhf()

    elif numpy.iscomplexobj(mo_coeff) or numpy.iscomplexobj(mf.mo_coeff):
        raise NotImplementedError

    else:
        return qcisd.QCISD(mf, frozen, mo_coeff, mo_occ)
RQCISD.__doc__ = qcisd.QCISD.__doc__



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Contributor(s):
# - Wirawan Purwanto
# - Qiming Sun <osirpt.sun@gmail.com>
#
#

import h5py
from pyscf.lib import H5FileWrap
from pyscf.lib.chkfile import load, load_mol, dump
from pyscf.mcscf.addons import StateAverageMixFCISolver


def load_mcscf(chkfile):
    return load_mol(chkfile), load(chkfile, "mcscf")


def dump_mcscf(
    mc,
    chkfile=None,
    key="mcscf",
    e_tot=None,
    mo_coeff=None,
    ncore=None,
    ncas=None,
    mo_occ=None,
    mo_energy=None,
    e_cas=None,
    ci_vector=None,
    casdm1=None,
    overwrite_mol=True,
):
    """Save CASCI/CASSCF calculation results or intermediates in chkfile."""
    if chkfile is None:
        chkfile = mc.chkfile
    if ncore is None:
        ncore = mc.ncore
    if ncas is None:
        ncas = mc.ncas
    if e_tot is None:
        e_tot = mc.e_tot
    if e_cas is None:
        e_cas = mc.e_cas
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff
    if mo_occ is None:
        mo_occ = mc.mo_occ
    # if ci_vector is None: ci_vector = mc.ci

    if h5py.is_hdf5(chkfile):
        mode = "a"
    else:
        mode = "w"

    mixed_ci = (
        isinstance(mc.fcisolver, StateAverageMixFCISolver) and ci_vector is not None
    )

    with H5FileWrap(chkfile, mode) as fh5:
        if mode == "a":
            if key in fh5:
                del fh5[key]

        if "mol" not in fh5:
            fh5["mol"] = mc.mol.dumps()
        elif overwrite_mol:
            del fh5["mol"]
            fh5["mol"] = mc.mol.dumps()

        def store(subkey, val):
            if val is not None:
                fh5[key + "/" + subkey] = val

        store("mo_coeff", mo_coeff)
        store("e_tot", e_tot)
        store("e_cas", e_cas)
        store("ncore", ncore)
        store("ncas", ncas)
        store("mo_occ", mo_occ)
        if mo_energy is not None:
            store("mo_energy", mo_energy)
        store("casdm1", casdm1)

        if not mixed_ci:
            store("ci", ci_vector)

    if mixed_ci:
        dump(chkfile, "mcscf/ci", ci_vector)



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Density functional theory
=========================

Simple usage::

    >>> from pyscf import gto, dft
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='def2-tzvp')
    >>> mf = dft.RKS(mol)
    >>> mf.xc = 'pbe,pbe'
    >>> mf.run()
'''

try:
    from pyscf.dft import libxc
    XC = {**libxc.XC, **libxc.XC_ALIAS}
except (ImportError, OSError):
    XC = None
try:
    from pyscf.dft import xcfun
    if XC is None:
        XC = {**xcfun.XC, **xcfun.XC_ALIAS}
except (ImportError, OSError):
    pass
from pyscf import gto
#from pyscf.dft import xc
from pyscf.dft import rks
from pyscf.dft import roks
from pyscf.dft import uks
from pyscf.dft import gks
from pyscf.dft import rks_symm
from pyscf.dft import uks_symm
from pyscf.dft import gks_symm
from pyscf.dft import dks
from pyscf.dft import gen_grid
grid = gen_grid
from pyscf.dft import radi
from pyscf.dft import numint
from pyscf.df import density_fit
from pyscf.dft.rks import KohnShamDFT
from pyscf.dft.gen_grid import sg1_prune, nwchem_prune, treutler_prune, \
        stratmann, original_becke, Grids
from pyscf.dft.radi import BRAGG_RADII, COVALENT_RADII, \
        delley, mura_knowles, gauss_chebyshev, treutler, treutler_ahlrichs, \
        treutler_atomic_radii_adjust, becke_atomic_radii_adjust


def KS(mol, xc='LDA,VWN'):
    if mol.spin == 0:
        return RKS(mol, xc)
    else:
        return UKS(mol, xc)
KS.__doc__ = '''
A wrap function to create DFT object (RKS or UKS).\n
''' + rks.RKS.__doc__
DFT = KS

def RKS(mol, xc='LDA,VWN'):
    if mol.spin == 0:
        if not mol.symmetry or mol.groupname == 'C1':
            return rks.RKS(mol, xc)
        else:
            return rks_symm.RKS(mol, xc)
    else:
        return ROKS(mol, xc)
RKS.__doc__ = rks.RKS.__doc__

def ROKS(mol, xc='LDA,VWN'):
    if not mol.symmetry or mol.groupname == 'C1':
        return roks.ROKS(mol, xc)
    else:
        return rks_symm.ROKS(mol, xc)
ROKS.__doc__ = roks.ROKS.__doc__

def UKS(mol, xc='LDA,VWN'):
    if not mol.symmetry or mol.groupname == 'C1':
        return uks.UKS(mol, xc)
    else:
        return uks_symm.UKS(mol, xc)
UKS.__doc__ = uks.UKS.__doc__

def GKS(mol, xc='LDA,VWN'):
    if not mol.symmetry or mol.groupname == 'C1':
        return gks.GKS(mol, xc)
    else:
        return gks_symm.GKS(mol, xc)
GKS.__doc__ = gks.GKS.__doc__

def DKS(mol, xc='LDA,VWN'):
    from pyscf.scf import dhf
    if dhf.zquatev and mol.spin == 0:
        return dks.RDKS(mol, xc=xc)
    else:
        return dks.UDKS(mol, xc=xc)
UDKS = dks.UDKS
RDKS = dks.RDKS

def X2C(mol, *args):
    '''X2C Kohn-Sham'''
    from pyscf.scf import dhf
    from pyscf.x2c import dft
    if dhf.zquatev and mol.spin == 0:
        return dft.RKS(mol, *args)
    else:
        return dft.UKS(mol, *args)
X2C_KS = X2C

def RKSpU(mol, xc='LDA,VWN', **kwargs):
    from pyscf.dft import rkspu
    return rkspu.RKSpU(mol, xc=xc, **kwargs)

def UKSpU(mol, xc='LDA,VWN', **kwargs):
    from pyscf.dft import ukspu
    return ukspu.UKSpU(mol, xc=xc, **kwargs)



#!/usr/bin/env python
#
# This code was copied from the data generation program of Tencent Alchemy
# project (https://github.com/tencent-alchemy).
#

#
# Copyright 2019 Tencent America LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#


import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad import rks as rks_grad
from pyscf.grad import uks as uks_grad
from pyscf.df.grad import rhf as df_rhf_grad


def get_veff(ks_grad, mol=None, dm=None):
    '''Coulomb + XC functional
    '''
    if mol is None: mol = ks_grad.mol
    if dm is None: dm = ks_grad.base.make_rdm1()
    t0 = (logger.process_clock(), logger.perf_counter())

    mf = ks_grad.base
    ni = mf._numint
    grids, nlcgrids = rks_grad._initialize_grids(ks_grad)

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, ks_grad.max_memory*.9-mem_now)
    if ks_grad.grid_response:
        exc, vxc = uks_grad.get_vxc_full_response(
                ni, mol, grids, mf.xc, dm,
                max_memory=max_memory, verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc_full_response(
                ni, mol, nlcgrids, xc, dm[0]+dm[1],
                max_memory=max_memory, verbose=ks_grad.verbose)
            exc += enlc
            vxc += vnlc
        logger.debug1(ks_grad, 'sum(grids response) %s', exc.sum(axis=0))
    else:
        exc, vxc = uks_grad.get_vxc(
                ni, mol, grids, mf.xc, dm,
                max_memory=max_memory, verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc(
                ni, mol, nlcgrids, xc, dm[0]+dm[1],
                max_memory=max_memory, verbose=ks_grad.verbose)
            vxc += vnlc
    t0 = logger.timer(ks_grad, 'vxc', *t0)

    if not ni.libxc.is_hybrid_xc(mf.xc):
        vj = ks_grad.get_j(mol, dm)
        vxc += vj[0] + vj[1]
        if ks_grad.auxbasis_response:
            e1_aux = vj.aux.sum ((0,1))
    else:
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=mol.spin)
        vj, vk = ks_grad.get_jk(mol, dm)
        if ks_grad.auxbasis_response:
            vk.aux = vk.aux * hyb
        vk[:] *= hyb # inplace * for vk[:] to keep the .aux tag
        if omega != 0:  # For range separated Coulomb operator
            vk_lr = ks_grad.get_k(mol, dm, omega=omega)
            vk[:] += vk_lr * (alpha - hyb)
            if ks_grad.auxbasis_response:
                vk.aux[:] += vk_lr.aux * (alpha - hyb)
        vxc += vj[0] + vj[1] - vk
        if ks_grad.auxbasis_response:
            e1_aux = vj.aux.sum ((0,1))
            e1_aux -= numpy.trace (vk.aux, axis1=0, axis2=1)

    if ks_grad.auxbasis_response:
        logger.debug1(ks_grad, 'sum(auxbasis response) %s', e1_aux.sum(axis=0))
        vxc = lib.tag_array(vxc, exc1_grid=exc, aux=e1_aux)
    else:
        vxc = lib.tag_array(vxc, exc1_grid=exc)
    return vxc


class Gradients(uks_grad.Gradients):

    _keys = {'with_df', 'auxbasis_response'}

    def __init__(self, mf):
        uks_grad.Gradients.__init__(self, mf)

    # Whether to include the response of DF auxiliary basis when computing
    # nuclear gradients of J/K matrices
    auxbasis_response = True
    get_jk = df_rhf_grad.Gradients.get_jk
    get_j = df_rhf_grad.Gradients.get_j
    get_k = df_rhf_grad.Gradients.get_k
    get_veff = get_veff

    def extra_force(self, atom_id, envs):
        e1 = uks_grad.Gradients.extra_force(self, atom_id, envs)
        if self.auxbasis_response:
            e1 += envs['vhf'].aux[atom_id]
        return e1

Grad = Gradients



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Parsers for basis set in the NWChem format
'''

__all__ = ['parse', 'load', 'parse_ecp', 'load_ecp',
           'convert_basis_to_nwchem', 'convert_ecp_to_nwchem',
           'optimize_contraction', 'remove_zero', 'to_general_contraction']

import re
import numpy
import numpy as np
import scipy.linalg
from pyscf.data.elements import _std_symbol
from pyscf.lib.exceptions import BasisNotFoundError
from pyscf.gto.basis.parse_nwchem_ecp import DISABLE_EVAL, MAXL, SPDF, MAPSPDF
from pyscf.gto.basis.parse_nwchem_ecp import parse as parse_ecp
from pyscf.gto.basis.parse_nwchem_ecp import load as load_ecp
from pyscf.gto.basis.parse_nwchem_ecp import convert_ecp_to_nwchem
from pyscf import __config__

BASIS_SET_DELIMITER = re.compile('# *BASIS SET.*\n|END\n')

def parse(string, symb=None, optimize=True):
    '''Parse the basis text which is in NWChem format. Return an internal
    basis format which can be assigned to attribute :attr:`Mole.basis`
    Empty lines, or the lines started with #, or the lines of "BASIS SET" and
    "END" will be ignored are ignored.

    Args:
        string : A string in NWChem basis format. Empty links and the lines of
        "BASIS SET" and "END" will be ignored

    Kwargs:
        optimize : Optimize basis contraction.  Convert the segment contracted
            basis to the general contracted basis.

    Examples:

    >>> mol = gto.Mole()
    >>> mol.basis = {'O': gto.basis.parse("""
    ... #BASIS SET: (6s,3p) -> [2s,1p]
    ... C    S
    ...      71.6168370              0.15432897
    ...      13.0450960              0.53532814
    ...       3.5305122              0.44463454
    ... C    SP
    ...       2.9412494             -0.09996723             0.15591627
    ...       0.6834831              0.39951283             0.60768372
    ...       0.2222899              0.70011547             0.39195739
    ... """)}

    >>> gto.basis.parse("""
    ... He    S
    ...      13.6267000              0.1752300
    ...       1.9993500              0.8934830
    ...       0.3829930              0.0000000
    ... He    S
    ...      13.6267000              0.0000000
    ...       1.9993500              0.0000000
    ...       0.3829930              1.0000000
    ... """, optimize=True)
    [[0, [13.6267, 0.17523, 0.0], [1.99935, 0.893483, 0.0], [0.382993, 0.0, 1.0]]]
    '''
    if symb is not None:
        symb = _std_symbol(symb)
        line_data = _search_basis_block(re.split(BASIS_SET_DELIMITER, string), symb)
        if not line_data:
            raise BasisNotFoundError('Basis set not found for %s' % symb)
    else:
        line_data = string.splitlines()

    raw_basis = []
    for dat in line_data:
        dat = dat.split('#')[0].strip()  # Use # to start comments
        dat_upper = dat.upper()
        if (dat and not dat_upper.startswith('END') and not dat_upper.startswith('BASIS')):
            raw_basis.append(dat)
    return _parse(raw_basis, optimize)

def load(basisfile, symb, optimize=True):
    '''Load basis for atom of symb from file'''
    raw_basis = search_seg(basisfile, symb)
    if not raw_basis:
        raise BasisNotFoundError('Basis set not found for %s' % symb)
    return _parse(raw_basis, optimize)

def _parse(raw_basis, optimize=True):
    basis_parsed = [[] for l in range(MAXL)]
    key = None
    for line in raw_basis:
        dat = line.strip()
        if not dat or dat.startswith('#'):
            continue
        elif dat[0].isalpha():
            keys = dat.split()
            if len(keys) == 1:
                key = keys[0].upper()
            else:
                key = keys[1].upper()
            if key == 'SP':
                basis_parsed[0].append([0])
                basis_parsed[1].append([1])
            elif key in MAPSPDF:
                l = MAPSPDF[key]
                current_basis = [l]
                basis_parsed[l].append(current_basis)
            else:
                raise BasisNotFoundError('Not basis data')
        else:
            dat = dat.replace('D','e').split()
            try:
                dat = [float(x) for x in dat]
            except ValueError:
                if DISABLE_EVAL:
                    raise ValueError('Failed to parse %s' % line)
                else:
                    dat = list(eval(','.join(dat)))
            except Exception as e:
                raise BasisNotFoundError('\n' + str(e) +
                                         '\nor the required basis file not existed.')
            if key is None:
                raise BasisNotFoundError('Not basis data')
            elif key == 'SP':
                basis_parsed[0][-1].append([dat[0], dat[1]])
                basis_parsed[1][-1].append([dat[0], dat[2]])
            else:
                current_basis.append(dat)
    basis_sorted = [b for bs in basis_parsed for b in bs]
    if not basis_sorted:
        raise BasisNotFoundError('Basis data not found')

    if optimize:
        basis_sorted = optimize_contraction(basis_sorted)

    basis_sorted = remove_zero(basis_sorted)
    return basis_sorted

def search_seg(basisfile, symb):
    symb = _std_symbol(symb)
    with open(basisfile, 'r') as fin:
        fdata = re.split(BASIS_SET_DELIMITER, fin.read())
    line_data = _search_basis_block(fdata, symb)
    return [x for x in line_data if x and 'END' not in x]

def _search_basis_block(raw_data, symb):
    line_data = []
    for dat in raw_data:
        dat0 = dat.split(None, 1)
        if not dat0:
            continue

        if dat0[0] == symb:
            return dat.splitlines()

        elif dat0[0][0] == '#':
            basis_lines = dat.splitlines()
            for i, line in enumerate(basis_lines):
                # Skip all leading '# xxx' lines and empty lines
                if not line or line.lstrip()[0] == '#':
                    continue
                elif line.split(None, 1)[0] == symb:
                    return basis_lines[i:]
                else:
                    break
    return line_data

def convert_basis_to_nwchem(symb, basis):
    '''Convert the internal basis format to NWChem format string'''
    res = []
    symb = _std_symbol(symb)

    # pass 1: comment line
    ls = [b[0] for b in basis]
    nprims = [len(b[1:]) for b in basis]
    nctrs = [len(b[1])-1 for b in basis]
    prim_to_ctr = {}
    for i, l in enumerate(ls):
        if l in prim_to_ctr:
            prim_to_ctr[l][0] += nprims[i]
            prim_to_ctr[l][1] += nctrs[i]
        else:
            prim_to_ctr[l] = [nprims[i], nctrs[i]]
    nprims = []
    nctrs = []
    for l in set(ls):
        nprims.append(str(prim_to_ctr[l][0])+SPDF[l].lower())
        nctrs.append(str(prim_to_ctr[l][1])+SPDF[l].lower())
    res.append('#BASIS SET: (%s) -> [%s]' % (','.join(nprims), ','.join(nctrs)))

    # pass 2: basis data
    for bas in basis:
        res.append('%-2s    %s' % (symb, SPDF[bas[0]]))
        for dat in bas[1:]:
            res.append(' '.join('%15.9f'%x for x in dat))
    return '\n'.join(res)

def optimize_contraction(basis):
    '''Search the basis segments which have the same exponents then merge them
    to the general contracted sets.

    Note the difference to the function :func:`to_general_contraction`. The
    return value of this function may still have multiple segments for each
    angular moment section.
    '''
    basdic = {}
    for b in basis:
        if isinstance(b[1], int):  # kappa = b[1]
            key = tuple(b[:2])
            ec = numpy.array(b[2:]).T
        else:
            key = tuple(b[:1])
            ec = numpy.array(b[1:]).T
        es = ec[0]
        cs = list(ec[1:])

        if key not in basdic:
            basdic[key] = []

        if basdic[key]:
            for e_cs in basdic[key]:
                if numpy.array_equal(e_cs[0], es):
                    e_cs.extend(cs)
                    break
            else:
                basdic[key].append([es] + cs)
        else:
            basdic[key].append([es] + cs)

    basis = []
    for key in sorted(basdic.keys()):
        l_kappa = list(key)
        for e_cs in basdic[key]:
            b = l_kappa + numpy.array(e_cs).T.tolist()
            basis.append(b)
    return basis

def to_general_contraction(basis):
    '''Segmented contracted basis -> general contracted basis.

    Combine multiple basis segments to one segment for each angular moment
    section.

    Examples:

    >>> gto.contract(gto.uncontract(gto.load('sto3g', 'He')))
    [[0, [6.36242139, 1.0, 0.0, 0.0], [1.158923, 0.0, 1.0, 0.0], [0.31364979, 0.0, 0.0, 1.0]]]
    '''
    basdic = {}
    for b in basis:
        if isinstance(b[1], int):  # kappa = b[1]
            key = tuple(b[:2])
            ec = numpy.array(b[2:])
        else:
            key = tuple(b[:1])
            ec = numpy.array(b[1:])
        if key in basdic:
            basdic[key].append(ec)
        else:
            basdic[key] = [ec]

    basis = []
    for key in sorted(basdic.keys()):
        l_kappa = list(key)

        es = numpy.hstack([ec[:,0] for ec in basdic[key]])
        cs = scipy.linalg.block_diag(*[ec[:,1:] for ec in basdic[key]])

        es, e_idx, rev_idx = numpy.unique(es.round(9), True, True)
        es = es[::-1]  # sort the exponents from large to small
        bcoeff = numpy.zeros((e_idx.size, cs.shape[1]))
        for i, j in enumerate(rev_idx):
            bcoeff[j] += cs[i]
        bcoeff = bcoeff[::-1]
        ec = numpy.hstack((es[:,None], bcoeff))

        basis.append(l_kappa + ec.tolist())

    return basis

def remove_zero(basis):
    '''
    Remove exponents if their contraction coefficients are all zeros.
    '''
    new_basis = []
    for b in basis:
        if isinstance(b[1], int):  # kappa = b[1]
            key = list(b[:2])
            ec = b[2:]
        else:
            key = list(b[:1])
            ec = b[1:]

        new_ec = [e_c for e_c in ec if any(c!=0 for c in e_c[1:])]
        if new_ec:
            new_basis.append(key + new_ec)
    return new_basis

if __name__ == '__main__':
    from pyscf import gto
    mol = gto.M(atom='O', basis='6-31g')
    print(load_ecp('lanl2dz.dat', 'Na'))
    b = load('ano.dat', 'Na')
    print(convert_basis_to_nwchem('Na', b))
    b = load_ecp('lanl2dz.dat', 'Na')
    print(convert_ecp_to_nwchem('Na', b))



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver Backhouse <olbackhouse@gmail.com>
#

"""
GF-CCSD solver via moment constraints.

See reference: Backhouse, Booth, arXiv:2206.13198 (2022).
"""

from collections import defaultdict

import numpy as np
import scipy.linalg

from pyscf import lib, cc, ao2mo
from pyscf.lib import logger
from pyscf.agf2 import mpi_helper


def kernel(
        gfccsd,
        hole_moments=None,
        part_moments=None,
        t1=None,
        t2=None,
        l1=None,
        l2=None,
        eris=None,
        imds=None,
        verbose=None,
):
    if gfccsd.verbose >= logger.WARN:
        gfccsd.check_sanity()
    gfccsd.dump_flags()

    log = logger.new_logger(gfccsd, verbose)

    if (l1 is None and gfccsd._cc.l1 is None) or (l2 is None and gfccsd._cc.l2 is None):
        raise ValueError(
                "Lambda amplitudes must be set for %s. This "
                "can be done by calling solve_lambda on the "
                "CC method or by setting l1, l2 attributes. "
                % gfccsd.__class__.__name__
        )

    if (hole_moments is None or part_moments is None) and imds is None:
        ip = hole_moments is None
        ea = part_moments is None
        imds = gfccsd.make_imds(eris=eris, ip=ip, ea=ea)

    if hole_moments is None:
        log.info("Building hole moments:")
        hole_moments = gfccsd.build_hole_moments(t1=t1, t2=t2, l1=l1, l2=l2, imds=imds)
    else:
        log.info("Hole moments passed by argument.")

    if part_moments is None:
        log.info("Building particle moments:")
        part_moments = gfccsd.build_part_moments(t1=t1, t2=t2, l1=l1, l2=l2, imds=imds)
    else:
        log.info("Particle moments passed by argument.")

    if gfccsd.hermi_moments:
        hole_moments = 0.5 * (hole_moments + hole_moments.swapaxes(1, 2).conj())
        part_moments = 0.5 * (part_moments + part_moments.swapaxes(1, 2).conj())

    if gfccsd.hermi_solver:
        solver = block_lanczos_symm
        eig = eigh_block_tridiagonal
    else:
        solver = block_lanczos_nosymm
        eig = eig_block_tridiagonal

    log.info("Solving for the hole moments.")
    blocks = solver(gfccsd, hole_moments)
    orth = mat_sqrt(hole_moments[0], hermi=gfccsd.hermi_solver)
    eh, vh = eig(gfccsd, *blocks, orth=orth)

    log.info("Solving for the particle moment.")
    blocks = solver(gfccsd, part_moments)
    orth = mat_sqrt(part_moments[0], hermi=gfccsd.hermi_solver)
    ep, vp = eig(gfccsd, *blocks, orth=orth)

    # Check the moments
    if gfccsd.niter[0] is not None:
        for n in range(2*gfccsd.niter[0]+2):
            a = lib.einsum("xk,yk,k->xy", vh[0], vh[1].conj(), eh**n)
            a /= np.max(np.abs(a))
            b = hole_moments[n] / np.max(np.abs(hole_moments[n]))
            err = np.max(np.abs(a - b))
            (logger.debug1 if err < 1e-8 else logger.warn)(
                    gfccsd, "Error in hole moment %d:  %10.6g", n, err)
    if gfccsd.niter[0] is not None:
        for n in range(2*gfccsd.niter[1]+2):
            a = lib.einsum("xk,yk,k->xy", vp[0], vp[1].conj(), ep**n)
            a /= np.max(np.abs(a))
            b = part_moments[n] / np.max(np.abs(part_moments[n]))
            err = np.max(np.abs(a - b))
            (logger.debug1 if err < 1e-8 else logger.warn)(
                    gfccsd, "Error in particle moment %d:  %10.6g", n, err)

    mask = np.argsort(eh.real)
    eh, vh = eh[mask], (vh[0][:, mask], vh[1][:, mask])
    mask = np.argsort(ep.real)
    ep, vp = ep[mask], (vp[0][:, mask], vp[1][:, mask])

    return eh, vh, ep, vp


def mat_sqrt(m, hermi=False):
    """Return the square root of a matrix.
    """

    if hermi:
        w, v = np.linalg.eigh(m)
        mask = w >= 0
        w, v = w[mask], v[:, mask]
        out = np.dot(v * w[None]**0.5, v.T.conj())

    else:
        w, v = np.linalg.eig(m)
        out = np.dot(v * w[None]**(0.5+0j), np.linalg.inv(v))

    return out


def mat_isqrt(m, tol=1e-16, hermi=False):
    """Return the inverse square root of a matrix.
    """

    if hermi:
        w, v = np.linalg.eigh(m)
        mask = w > tol
        w, v = w[mask], v[:, mask]
        out = np.dot(v * w[None]**-0.5, v.T.conj())

    else:
        w, v = np.linalg.eig(m)
        mask = np.abs(w) >= tol
        vinv = np.linalg.inv(v)[mask]
        w, v = w[mask], v[:, mask]
        out = np.dot(v * w[None]**(-0.5+0j), vinv)

    return out


def build_block_tridiagonal(a, b, c=None):
    """Construct a block tridiagonal matrix from a list of on-diagonal
    and off-diagonal blocks.
    """

    z = np.zeros_like(a[0], dtype=a[0].dtype)

    if c is None:
        c = [x.T.conj() for x in b]

    h = np.block([[
        a[i] if i == j else
        b[j] if j == i-1 else
        c[i] if i == j-1 else z
        for j in range(len(a))]
        for i in range(len(a))]
    )

    return h


def eig_block_tridiagonal(gfccsd, a, b, c, orth=None):
    """Diagonalise a non-Hermitian block-tridiagonal Hamiltonian and
    transform its eigenvectors appropriately.
    """

    h_tri = build_block_tridiagonal(a, b, c)

    e, u = np.linalg.eig(h_tri)

    if orth is not None:
        vl = np.dot(orth, u[:gfccsd.nmo])
        vr = np.dot(np.linalg.inv(u)[:, :gfccsd.nmo], orth).T.conj()
    else:
        vl = u[:gfccsd.nmo]
        vr = np.linalg.inv(u)[:, :gfccsd.nmo].T.conj()

    return e, (vl, vr)


def eigh_block_tridiagonal(gfccsd, a, b, orth=None):
    """Diagonalise a Hermitian block-tridiagonal Hamiltonian and
    transform its eigenvectors appropriately.
    """

    h_tri = build_block_tridiagonal(a, b)

    e, u = np.linalg.eigh(h_tri)

    if orth is not None:
        v = np.dot(orth, u[:gfccsd.nmo])
    else:
        v = u[:gfccsd.nmo]

    return e, (v, v)


def _matrix_info(x, hermi=False):
    norm = np.abs(np.einsum("pq,qp->", x, x))
    eigvals = np.linalg.eigvals(x)
    mineig = np.min(np.abs(eigvals))
    maxeig = np.max(np.abs(eigvals))
    return norm, mineig, maxeig


def block_lanczos_symm(gfccsd, moments, verbose=None):
    """Hermitian block Lanczos solver, returns a set of poles that
    best reproduce the inputted moments.

    Args:
        gfccsd : MomGFCCSD
            GF-CCSD object
        moments : ndarray (2*niter+2, n, n)
            Array of moments with which the resulting poles should
            be consistent with.

    Kwargs:
        verbose : int
            Level of verbosity.

    Returns:
        a : ndarray (niter+1, n, n)
            On-diagonal blocks of the block tridiagonal Hamiltonian.
        b : ndarray (niter, n, n)
            Off-diagonal blocks of the block tridiagonal Hamiltonian.
    """

    log = logger.new_logger(gfccsd, verbose)
    log.debug1("block_lanczos_symm: %d moments", len(moments))

    nmo = gfccsd.nmo
    niter = (len(moments) - 2) // 2
    dtype = np.complex128

    a = np.zeros((niter+1, nmo, nmo), dtype=dtype)
    b = np.zeros((niter, nmo, nmo), dtype=dtype)
    t = np.zeros((len(moments), nmo, nmo), dtype=dtype)

    v = defaultdict(lambda: np.zeros((nmo, nmo), dtype=dtype))
    v[0, 0] = np.eye(nmo).astype(dtype)

    orth = mat_isqrt(moments[0], hermi=True)
    for i in range(len(moments)):
        t[i] = np.linalg.multi_dot((orth, moments[i], orth))

    a[0] = t[1]

    log.debug1("Raw moments:")
    log.debug1("  %4s %12s %12s %12s", "N", "norm", "min(|eig|)", "max(|eig|)")
    for i in range(len(moments)):
        log.debug1("  %4d %12.6g %12.6g %12.6g", i, *_matrix_info(moments[i], hermi=True))

    log.debug1("Orthogonalised moments:")
    log.debug1("  %4s %12s %12s %12s", "N", "norm", "min(|eig|)", "max(|eig|)")
    for i in range(len(moments)):
        log.debug1("  %4d %12.6g %12.6g %12.6g", i, *_matrix_info(t[i], hermi=True))

    for i in range(niter):
        log.info("Iteration %d", i)

        b2 = np.zeros((nmo, nmo), dtype=dtype)
        for j in range(i+2):
            for l in range(i+1):
                b2 += np.linalg.multi_dot((v[i, l].T.conj(), t[j+l+1], v[i, j-1]))

        b2 -= np.dot(a[i], a[i])
        if i:
            b2 -= np.dot(b[i-1], b[i-1])

        b[i] = mat_sqrt(b2, hermi=True)
        binv = mat_isqrt(b2, hermi=True)

        for j in range(i+2):
            r = (
                    + v[i, j-1]
                    - np.dot(v[i, j], a[i])
                    - np.dot(v[i-1, j], b[i-1])
            )
            v[i+1, j] = np.dot(r, binv)

        for j in range(i+2):
            for l in range(i+2):
                a[i+1] += np.linalg.multi_dot((v[i+1, l].T.conj(), t[j+l+1], v[i+1, j]))

        log.debug1("  %4s %12s %12s %12s", "mat", "norm", "min(|eig|)", "max(|eig|)")
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B^2", *_matrix_info(b2))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B", *_matrix_info(b[i]))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B^-1", *_matrix_info(binv))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "A", *_matrix_info(a[i+1]))

        biorth_error = 0.0
        for j in range(i+2):
            x = np.zeros_like(v[0, 0])
            for k in range(i+2):
                for l in range(i+2):
                    x += np.linalg.multi_dot((v[i+1, l].T.conj(), t[k+l], v[j, k]))
            biorth_error = max(biorth_error, np.max(np.abs(x - np.eye(nmo)*((i+1)==j))))
        log.info("  Error in biorthogonality:  %12.6g", biorth_error)

    return a, b


def block_lanczos_nosymm(gfccsd, moments, verbose=None):
    """Non-Hermitian block Lanczos solver, returns a set of poles that
    best reproduce the inputted moments.

    Args:
        gfccsd : MomGFCCSD
            GF-CCSD object
        moments : ndarray (2*niter+2, n, n)
            Array of moments with which the resulting poles should
            be consistent with.

    Kwargs:
        verbose : int
            Level of verbosity.

    Returns:
        a : ndarray (niter+1, n, n)
            On-diagonal blocks of the block tridiagonal Hamiltonian.
        b : ndarray (niter, n, n)
            Upper off-diagonal blocks of the block tridiagonal
            Hamiltonian.
        c : ndarray (niter, n, n)
            Lower off-diagonal blocks of the block tridiagonal
            Hamiltonian.
    """

    log = logger.new_logger(gfccsd, verbose)
    log.debug1("block_lanczos_nosymm: %d moments", len(moments))

    nmo = gfccsd.nmo
    niter = (len(moments) - 2) // 2
    dtype = np.complex128

    a = np.zeros((niter+1, nmo, nmo), dtype=dtype)
    b = np.zeros((niter, nmo, nmo), dtype=dtype)
    c = np.zeros((niter, nmo, nmo), dtype=dtype)
    t = np.zeros((len(moments), nmo, nmo), dtype=dtype)

    v = defaultdict(lambda: np.zeros((nmo, nmo), dtype=dtype))
    w = defaultdict(lambda: np.zeros((nmo, nmo), dtype=dtype))
    v[0, 0] = np.eye(nmo).astype(dtype)
    w[0, 0] = np.eye(nmo).astype(dtype)

    orth = mat_isqrt(moments[0])
    for i in range(len(moments)):
        t[i] = np.linalg.multi_dot((orth, moments[i], orth))

    a[0] = t[1]

    log.debug1("Raw moments:")
    log.debug1("  %4s %12s %12s %12s", "N", "norm", "min(|eig|)", "max(|eig|)")
    for i in range(len(moments)):
        log.debug1("  %4d %12.6g %12.6g %12.6g", i, *_matrix_info(moments[i]))

    log.debug1("Orthogonalised moments:")
    log.debug1("  %4s %12s %12s %12s", "N", "norm", "min(|eig|)", "max(|eig|)")
    for i in range(len(moments)):
        log.debug1("  %4d %12.6g %12.6g %12.6g", i, *_matrix_info(t[i]))

    for i in range(niter):
        log.info("Iteration %d", i)

        b2 = np.zeros((nmo, nmo), dtype=dtype)
        c2 = np.zeros((nmo, nmo), dtype=dtype)

        for j in range(i+2):
            for l in range(i+1):
                b2 += np.linalg.multi_dot((w[i, l], t[j+l+1], v[i, j-1]))
                c2 += np.linalg.multi_dot((w[i, j-1], t[j+l+1], v[i, l]))

        b2 -= np.dot(a[i], a[i])
        c2 -= np.dot(a[i], a[i])
        if i:
            b2 -= np.dot(c[i-1], c[i-1])
            c2 -= np.dot(b[i-1], b[i-1])

        b[i] = mat_sqrt(b2)
        c[i] = mat_sqrt(c2)

        binv = mat_isqrt(b2)
        cinv = mat_isqrt(c2)

        for j in range(i+2):
            r = (
                    + v[i, j-1]
                    - np.dot(v[i, j], a[i])
                    - np.dot(v[i-1, j], b[i-1])
            )
            v[i+1, j] = np.dot(r, cinv)

            s = (
                    + w[i, j-1]
                    - np.dot(a[i], w[i, j])
                    - np.dot(c[i-1], w[i-1, j])
            )
            w[i+1, j] = np.dot(binv, s)

        for j in range(i+2):
            for l in range(i+2):
                a[i+1] += np.linalg.multi_dot((w[i+1, l], t[j+l+1], v[i+1, j]))

        log.debug1("  %4s %12s %12s %12s", "mat", "norm", "min(|eig|)", "max(|eig|)")
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B^2", *_matrix_info(b2))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B", *_matrix_info(b[i]))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "B^-1", *_matrix_info(binv))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "C^2", *_matrix_info(c2))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "C", *_matrix_info(c[i]))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "C^-1", *_matrix_info(cinv))
        log.debug1("  %4s %12.6g %12.6g %12.6g", "A", *_matrix_info(a[i+1]))

        biorth_error = 0.0
        for j in range(i+2):
            x = np.zeros_like(v[0, 0])
            y = np.zeros_like(v[0, 0])
            for k in range(i+2):
                for l in range(i+2):
                    x += np.linalg.multi_dot((w[i+1, l], t[k+l], v[j, k]))
                    y += np.linalg.multi_dot((w[j, l], t[k+l], v[i+1, k]))
            biorth_error = max(biorth_error, np.max(np.abs(x - np.eye(nmo)*((i+1)==j))))
            biorth_error = max(biorth_error, np.max(np.abs(y - np.eye(nmo)*((i+1)==j))))
        log.info("  Error in biorthogonality:  %12.6g", biorth_error)

    return a, b, c


def _kd(n, i):
    v = np.zeros((n,))
    v[i] = 1.0
    return v


def contract_ket_hole(gfccsd, eom, t1, t2, v, orb):
    r"""Contract a vector with \bar{a}^\dagger_p |\Psi>.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        return v[orb]
    else:
        b1 = t1[:, orb-nocc]
        b2 = t2[:, :, orb-nocc]
        b = eom.amplitudes_to_vector(b1, b2)
        return np.dot(v, b)


def build_ket_hole(gfccsd, eom, t1, t2, orb):
    r"""Build \bar{a}^\dagger_p |\Psi>.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        b1 = np.eye(nocc)[orb]
        b2 = np.zeros((nocc, nocc, nvir))
    else:
        b1 = t1[:, orb-nocc]
        b2 = t2[:, :, orb-nocc]

    return eom.amplitudes_to_vector(b1, b2)


def build_bra_hole(gfccsd, eom, t1, t2, l1, l2, orb):
    """Get the first- and second-order contributions to the left-hand
    transformed vector for a given orbital for the hole part of the
    Green's function.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        e1 = _kd(nocc, orb)
        e1 -= lib.einsum("ie,e->i", l1, t1[orb])
        tmp = t2[orb] * 2.0
        tmp -= t2[orb].swapaxes(1, 2)
        e1 -= lib.einsum("imef,mef->i", l2, tmp)

        tmp = -lib.einsum("ijea,e->ija", l2, t1[orb])
        e2 = 2.0 * tmp
        e2 -= tmp.swapaxes(0, 1)
        tmp = lib.einsum("ja,i->ija", l1, _kd(nocc, orb))
        e2 += tmp * 2.0
        e2 -= tmp.swapaxes(0, 1)

    else:
        e1 = l1[:, orb-nocc].copy()
        e2 = l2[:, :, orb-nocc] * 2.0
        e2 -= l2[:, :, :, orb-nocc]

    return eom.amplitudes_to_vector(e1, e2)


def contract_ket_part(gfccsd, eom, t1, t2, v, orb):
    r"""Contract a vector with \bar{a}_p |\Psi>.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        b1 = t1[orb]
        b2 = t2[orb]
        b = eom.amplitudes_to_vector(b1, b2)
        return np.dot(v, b)
    else:
        return -v[orb-nocc]


def build_ket_part(gfccsd, eom, t1, t2, orb):
    r"""Build \bar{a}_p |\Psi>.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        b1 = t1[orb]
        b2 = t2[orb]
    else:
        b1 = -np.eye(nvir)[orb-nocc]
        b2 = np.zeros((nocc, nvir, nvir))

    return eom.amplitudes_to_vector(b1, b2)


def build_bra_part(gfccsd, eom, t1, t2, l1, l2, orb):
    """Get the first- and second-order contributions to the left-hand
    transformed vector for a given orbital for the particle part of the
    Green's function.
    """

    nocc, nvir = t1.shape

    if orb < nocc:
        e1 = -l1[orb]
        e2 = -l2[orb] * 2.0
        e2 += l2[:, orb]

    else:
        e1 = _kd(nvir, orb-nocc)
        e1 -= lib.einsum("mb,m->b", l1, t1[:, orb-nocc])
        tmp = t2[:, :, :, orb-nocc] * 2.0
        tmp -= t2[:, :, orb-nocc]
        e1 -= lib.einsum("kmeb,kme->b", l2, tmp)

        tmp = -lib.einsum("ikba,k->iab", l2, t1[:, orb-nocc])
        e2 = tmp * 2.0
        e2 -= tmp.swapaxes(1, 2)
        tmp = lib.einsum("ib,a->iab", l1, _kd(nvir, orb-nocc))
        e2 += tmp * 2.0
        e2 -= tmp.swapaxes(1, 2)

    return eom.amplitudes_to_vector(e1, e2)


class MomGFCCSD(lib.StreamObject):
    """Green's function coupled cluster singles and doubles using the
    moment-resolved solver.

    Attributes:
        verbose : int
            Print level. Default value equals to :class:`Mole.verbose`.
        niter : tuple of (int, int)
            Number of block Lanczos iterations for occupied and virtual
            sectors. If either are `None` then said sector will not be
            computed.
        weight_tol : float
            Threshold for weight in the physical space to consider a
            pole an ionisation or removal event. Default value is 1e-1.
        hermi_moments : bool
            Whether to Hermitise the moments, default value is False.
        hermi_solver : obol
            Whether to use the real-valued, symmetric block Lanczos
            solver, default value is False.

    Results:
        eh : ndarray
            Energies of the compressed hole Green's function
        vh : tuple of ndarray
            Left- and right-hand transition amplitudes of the compressed
            hole Green's function
        ep : ndarray
            Energies of the compressed particle Green's function
        vp : tuple of ndarray
            Left- and right-hand transition amplitudes of the compressed
            particle Green's function
    """

    _keys = {
        'verbose', 'stdout', 'niter', 'weight_tol',
        'hermi_moments', 'hermi_solver', 'eh', 'ep', 'vh', 'vp', 'chkfile',
    }

    def __init__(self, mycc, niter=(2, 2)):
        self._cc = mycc
        self.verbose = mycc.verbose
        self.stdout = mycc.stdout

        if isinstance(mycc, cc.uccsd.UCCSD):
            raise NotImplementedError("MomGFCCSD for unrestricted CCSD")

        if isinstance(niter, int):
            self.niter = (niter, niter)
        else:
            self.niter = niter
        self.weight_tol = 1e-1
        self.hermi_moments = False
        self.hermi_solver = False
        self.eh = None
        self.ep = None
        self.vh = None
        self.vp = None
        self._t1 = None
        self._t2 = None
        self._l1 = None
        self._l2 = None
        self.chkfile = self._cc.chkfile

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info("")
        log.info("******** %s ********", self.__class__)
        log.info("niter = %s", self.niter)
        log.info("nmo = %s", self.nmo)
        log.info("nocc = %s", self.nocc)
        log.info("weight_tol = %s", self.weight_tol)
        log.info("hermi_moments = %s", self.hermi_moments)
        log.info("hermi_solver = %s", self.hermi_solver)
        log.info("chkfile = %s", self.chkfile)

    def _finalize(self):
        self.ipgfccsd()
        self.eagfccsd()
        return self

    def reset(self, mol=None):
        self._cc.reset(mol)
        return self

    @property
    def eomip_method(self):
        return self._cc.eomip_method()

    @property
    def eomea_method(self):
        return self._cc.eomea_method()

    build_bra_hole = build_bra_hole
    build_bra_part = build_bra_part
    contract_ket_hole = contract_ket_hole
    contract_ket_part = contract_ket_part

    def make_imds(self, eris=None, ip=True, ea=True):
        """Build EOM intermediates.
        """

        imds = cc.eom_rccsd._IMDS(self._cc, eris=eris)

        if ip:
            imds.make_ip()
        if ea:
            imds.make_ea()

        return imds

    def build_hole_moments(self, t1=None, t2=None, l1=None, l2=None, imds=None, niter=None):
        """Build moments of the hole (IP-EOM-CCSD) Green's function.
        """

        if t1 is None:
            t1 = self._cc.t1
        if t2 is None:
            t2 = self._cc.t2
        if l1 is None:
            l1 = self._cc.l1
        if l2 is None:
            l2 = self._cc.l2

        if niter is None:
            niter = self.niter[0]
        nmom = 2 * niter + 2
        moments = np.zeros((nmom, self.nmo, self.nmo))

        cput0 = (logger.process_clock(), logger.perf_counter())

        eom = self.eomip_method()
        if imds is None:
            imds = self.make_imds(ea=False)
        diag = eom.get_diag(imds)

        for p in mpi_helper.nrange(self.nmo):
            ket = self.build_bra_hole(eom, t1, t2, l1, l2, p)
            for n in range(nmom):
                for q in range(self.nmo):
                    moments[n, q, p] += self.contract_ket_hole(eom, t1, t2, ket, q)
                if (n+1) != nmom:
                    ket = -eom.l_matvec(ket, imds, diag)

        mpi_helper.barrier()
        moments = mpi_helper.allreduce(moments)

        logger.timer(self, "IP-EOM-CCSD moments", *cput0)

        return moments

    def build_part_moments(self, t1=None, t2=None, l1=None, l2=None, imds=None, niter=None):
        """Build moments of the particle (EA-EOM-CCSD) Green's function.
        """

        if t1 is None:
            t1 = self._cc.t1
        if t2 is None:
            t2 = self._cc.t2
        if l1 is None:
            l1 = self._cc.l1
        if l2 is None:
            l2 = self._cc.l2

        if niter is None:
            niter = self.niter[1]
        nmom = 2 * niter + 2
        moments = np.zeros((nmom, self.nmo, self.nmo))

        cput0 = (logger.process_clock(), logger.perf_counter())

        eom = self.eomea_method()
        if imds is None:
            imds = self.make_imds(ip=False)
        diag = eom.get_diag(imds)

        for p in mpi_helper.nrange(self.nmo):
            ket = self.build_bra_part(eom, t1, t2, l1, l2, p)
            for n in range(nmom):
                for q in range(self.nmo):
                    moments[n, q, p] -= self.contract_ket_part(eom, t1, t2, ket, q)
                if (n+1) != nmom:
                    ket = eom.l_matvec(ket, imds, diag)

        mpi_helper.barrier()
        moments = mpi_helper.allreduce(moments)

        logger.timer(self, "EA-EOM-CCSD moments", *cput0)

        return moments

    def make_rdm1(self, ao_repr=False, eris=None, imds=None):
        """Build the first-order reduced density matrix at the CCSD
        level using the zeroth-order moment of the hole part of the
        CCSD Green's function.
        """

        if imds is None:
            imds = self.make_imds(eris=eris, ea=False)

        dm1 = self.build_hole_moments(imds=imds, niter=0)[0]
        dm1 = dm1 + dm1.T.conj()

        if ao_repr:
            mo = self._cc.mo_coeff
            dm1 = np.linalg.multi_dot((mo, dm1, mo.T.conj()))

        return dm1

    def kernel(self, **kwargs):
        eh, vh, ep, vp = kernel(self, **kwargs)

        self.eh = eh
        self.vh = vh
        self.ep = ep
        self.vp = vp

        self._finalize()

        return eh, vh, ep, vp

    def dump_chk(self, chkfile=None, key="gfccsd"):
        if chkfile is None:
            chkfile = self.chkfile

        lib.chkfile.dump(chkfile, key+"/eh", self.eh)
        lib.chkfile.dump(chkfile, key+"/vh_left", self.vh[0])
        lib.chkfile.dump(chkfile, key+"/vh_right", self.vh[1])
        lib.chkfile.dump(chkfile, key+"/ep", self.ep)
        lib.chkfile.dump(chkfile, key+"/vp_left", self.vp[0])
        lib.chkfile.dump(chkfile, key+"/vp_right", self.vp[1])
        lib.chkfile.dump(chkfile, key+"/niter", np.array(self.niter))

        return self

    def update_from_chk_(self, chkfile=None, key="gfccsd"):
        if chkfile is None:
            chkfile = self.chkfile

        self.eh = lib.chkfile.load(chkfile, key+"/eh")
        self.vh = (
                lib.chkfile.load(chkfile, key+"/vh_left"),
                lib.chkfile.load(chkfile, key+"/vh_right"),
        )
        self.ep = lib.chkfile.load(chkfile, key+"/ep")
        self.vp = (
                lib.chkfile.load(chkfile, key+"/vp_left"),
                lib.chkfile.load(chkfile, key+"/vp_right"),
        )
        self.niter = tuple(lib.chkfile.load(chkfile, key+"/niter"))

    update = update_from_chk = update_from_chk_

    def ipgfccsd(self, nroots=5):
        """Print and return ionisation potentials.
        """

        eh, (vh, uh) = self.eh, self.vh

        mask = np.abs(np.sum(vh * uh.conj(), axis=0)) > self.weight_tol
        mask = np.arange(mask.size)[mask][::-1]
        e_ip = -eh[mask]
        v_ip, u_ip = vh[:, mask], uh[:, mask]

        nroots = min(nroots, len(e_ip))
        logger.note(self, "  %s %s %16s %10s", "", "", "Energy", "Weight")
        for n in range(nroots):
            qpwt = np.abs(np.sum(v_ip[:, n] * u_ip[:, n].conj())).real
            warn = ""
            if np.abs(e_ip[n].imag) > 1e-8:
                warn += "(Warning: imag part: %.6g)" % e_ip[n].imag
            logger.note(self, "  %2s %2d %16.10g %10.6g %s" % ("IP", n, e_ip[n].real, qpwt, warn))

        if nroots == 1:
            return e_ip[0].real, v_ip[:, 0], u_ip[:, 0]
        else:
            return e_ip.real, v_ip, u_ip

    def eagfccsd(self, nroots=5):
        """Print and return electron affinities.
        """

        ep, (vp, up) = self.ep, self.vp

        mask = np.abs(np.sum(vp * up.conj(), axis=0)) > self.weight_tol
        e_ea = ep[mask]
        v_ea, u_ea = vp[:, mask], up[:, mask]

        nroots = min(nroots, len(e_ea))
        logger.note(self, "  %s %s %16s %10s", "", "", "Energy", "Weight")
        for n in range(nroots):
            qpwt = np.abs(np.sum(v_ea[:, n] * u_ea[:, n].conj())).real
            warn = ""
            if np.abs(e_ea[n].imag) > 1e-8:
                warn += "(Warning: imag part: %.6g)" % e_ea[n].imag
            logger.note(self, "  %2s %2d %16.10g %10.6g %s" % ("EA", n, e_ea[n].real, qpwt, warn))

        if nroots == 1:
            return e_ea[0].real, v_ea[:, 0], u_ea[:, 0]
        else:
            return e_ea.real, v_ea, u_ea

    @property
    def nmo(self):
        return self._cc.nmo

    @property
    def nocc(self):
        return self._cc.nocc


if __name__ == "__main__":
    from pyscf import gto, scf

    mol = gto.M(
            #atom="O 0 0 0; O 0 0 1",
            atom="N 0 0 0; N 0 0 1",
            basis="cc-pvdz",
            verbose=0,
    )
    mf = scf.RHF(mol)
    mf = mf.run()
    ccsd = cc.CCSD(mf)
    ccsd = ccsd.run()
    ccsd.solve_lambda()

    niter = 5

    gfcc = MomGFCCSD(ccsd, (niter, niter))
    gfcc.kernel()

    ip1, vip1 = ccsd.ipccsd(nroots=8)
    ip2, vip2, uip2 = gfcc.ipgfccsd(nroots=8)

    ea1, vea1 = ccsd.eaccsd(nroots=8)
    ea2, vea2, uea2 = gfcc.eagfccsd(nroots=8)

    print("    %12s %12s %12s" % ("EOM", "GF", "Error"))
    print("IP1 %12.8f %12.8f %12.8f" % (ip1[0],ip2[0],np.abs(ip1[0]-ip2[0])))
    print("IP2 %12.8f %12.8f %12.8f" % (ip1[1],ip2[1],np.abs(ip1[1]-ip2[1])))
    print("EA1 %12.8f %12.8f %12.8f" % (ea1[0],ea2[0],np.abs(ea1[0]-ea2[0])))
    print("EA2 %12.8f %12.8f %12.8f" % (ea1[1],ea2[1],np.abs(ea1[1]-ea2[1])))



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''Analytic GTH-PP integrals for open boundary conditions.

See also pyscf/pbc/gto/pseudo/pp_int.py
'''

import numpy as np
from pyscf import lib
from pyscf.gto.mole import ATOM_OF, intor_cross

def get_gth_pp(mol):
    from pyscf.pbc.gto.pseudo import pp_int
    from pyscf.df import incore

    # Analytical integration for get_pp_loc_part1(cell).
    fakemol = pp_int.fake_cell_vloc(mol, 0)
    vpploc = 0
    if fakemol.nbas > 0:
        charges = fakemol.atom_charges()
        atmlst = fakemol._bas[:,ATOM_OF]
        v = incore.aux_e2(mol, fakemol, 'int3c2e', aosym='s2', comp=1)
        vpploc = np.einsum('...i,i->...', v, -charges[atmlst])

    intors = ('int3c2e', 'int3c1e', 'int3c1e_r2_origk',
              'int3c1e_r4_origk', 'int3c1e_r6_origk')
    for cn in range(1, 5):
        fakemol = pp_int.fake_cell_vloc(mol, cn)
        if fakemol.nbas > 0:
            v = incore.aux_e2(mol, fakemol, intors[cn], aosym='s2', comp=1)
            vpploc += np.einsum('...i->...', v)

    if isinstance(vpploc, np.ndarray):
        vpploc = lib.unpack_tril(vpploc)

    fakemol, hl_blocks = pp_int.fake_cell_vnl(mol)
    hl_dims = np.array([len(hl) for hl in hl_blocks])
    _bas = fakemol._bas
    ppnl_half = []
    intors = ('int1e_ovlp', 'int1e_r2_origi', 'int1e_r4_origi')
    for i, intor in enumerate(intors):
        fakemol._bas = _bas[hl_dims>i]
        if fakemol.nbas > 0:
            ppnl_half.append(intor_cross(intor, fakemol, mol))
        else:
            ppnl_half.append(None)
    fakemol._bas = _bas

    nao = mol.nao
    offset = [0] * 3
    for ib, hl in enumerate(hl_blocks):
        l = fakemol.bas_angular(ib)
        nd = 2 * l + 1
        hl_dim = hl.shape[0]
        ilp = np.empty((hl_dim, nd, nao))
        for i in range(hl_dim):
            p0 = offset[i]
            if ppnl_half[i] is None:
                ilp[i] = 0.
            else:
                ilp[i] = ppnl_half[i][p0:p0+nd]
            offset[i] = p0 + nd
        vpploc += np.einsum('ilp,ij,jlq->pq', ilp.conj(), hl, ilp)
    return vpploc
